# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Memory leak test

CodraFT application memory leak test.
"""

import os

import numpy as np
import psutil

from codraft.env import execenv
from codraft.tests.embedded1_unit import HostWindow
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_curves

SHOW = True  # Show test in GUI-based test launcher


def memory_leak_test(iterations=100):
    """Test for memory leak"""
    with qt_app_context():
        proc = psutil.Process(os.getpid())
        mainview = HostWindow()
        mainview.show()
        memlist = []
        for i in range(iterations):
            mainview.open_codraft()
            mainview.codraft.close()
            # mainview.codraft.destroy()
            # mainview.codraft = None
            # QApplication.processEvents()
            # import time; time.sleep(2)
            # QApplication.processEvents()
            memdata = proc.memory_info().vms / 1024**2
            memlist.append(memdata)
            execenv.print(i + 1, ":", memdata, "MB")
        view_curves(
            np.array(memlist),
            title="Memory leak test for CodraFT application",
            ylabel="Memory (MB)",
        )


if __name__ == "__main__":
    memory_leak_test()



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Codra
# Pierre Raybaut

"""
Module providing a log viewer widget, a log viewer window and CodraFT's log viewer
"""

import os.path as osp
from pathlib import Path

from guidata.configtools import get_icon
from guidata.widgets.codeeditor import CodeEditor
from qtpy import QtWidgets as QW

from codraft.config import APP_NAME, Conf, _, get_old_log_fname
from codraft.env import execenv
from codraft.utils.qthelpers import exec_dialog


def get_title_contents(path):
    """Get title and contents for log filename"""
    with open(path, "r", encoding="utf-8") as fdesc:
        contents = fdesc.read()
    pathobj = Path(path)
    uri_path = pathobj.absolute().as_uri()
    text = f'{_("Contents of file")} <a href="{uri_path}">{path}</a>:'
    return text, contents


class LogViewerWidget(QW.QWidget):
    """Log viewer widget"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.editor = CodeEditor(language="Python")
        self.editor.setReadOnly(True)
        layout = QW.QVBoxLayout()
        self.label = QW.QLabel("")
        layout.addWidget(self.label)
        layout.addWidget(self.editor)
        self.setLayout(layout)

    def set_data(self, text, contents):
        """Set log data"""
        self.label.setText(text)
        self.label.setOpenExternalLinks(True)
        self.editor.setPlainText(contents)


class LogViewerWindow(QW.QDialog):
    """Log viewer window"""

    def __init__(self, fnames, parent=None):
        super().__init__(parent)
        self.setObjectName("logviewer")
        self.setWindowTitle(_("CodraFT log files"))
        self.setWindowIcon(get_icon("codraft.svg"))
        self.tabs = QW.QTabWidget(self)
        for fname in fnames:
            if osp.isfile(fname):
                viewer = LogViewerWidget()
                title, contents = get_title_contents(fname)
                if not contents.strip():
                    continue
                viewer.set_data(title, contents)
                self.tabs.addTab(viewer, get_icon("logs.svg"), osp.basename(fname))
        layout = QW.QVBoxLayout()
        layout.addWidget(self.tabs)
        self.setLayout(layout)
        self.resize(900, 400)

    @property
    def is_empty(self):
        """Return True if there is no log available"""
        return self.tabs.count() == 0


def exec_codraft_logviewer_dialog(parent=None):
    """View CodraFT logs"""
    fnames = [
        osp.normpath(fname)
        for fname in (
            Conf.main.traceback_log_path.get(),
            Conf.main.faulthandler_log_path.get(),
            get_old_log_fname(Conf.main.traceback_log_path.get()),
            get_old_log_fname(Conf.main.faulthandler_log_path.get()),
        )
        if osp.isfile(fname)
    ]
    dlg = LogViewerWindow(fnames, parent=parent)
    if dlg.is_empty:
        if not execenv.unattended:
            QW.QMessageBox.information(
                dlg, APP_NAME, _("Log files are currently empty.")
            )
        dlg.close()
    else:
        exec_dialog(dlg)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
ROI test:

  - Defining Region of Interest on a signal
  - Defining Region of Interest on an image
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np

from codraft.core.gui.panel import ImagePanel, SignalPanel
from codraft.core.gui.processor.image import PeakDetectionParam
from codraft.core.gui.processor.signal import FWHMParam
from codraft.core.model.signal import SignalParam
from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.tests.data import create_test_image3, create_test_signal1

SHOW = True  # Show test in GUI-based test launcher


def test_signal_features(panel: SignalPanel, singleobj: bool = None):
    """Test all signal features related to ROI"""
    panel.processor.compute_fwhm(FWHMParam())
    panel.processor.compute_fw1e2()
    panel.processor.extract_roi(singleobj=singleobj)


def test_image_features(panel: ImagePanel, singleobj: bool = None):
    """Test all image features related to ROI"""
    panel.processor.compute_centroid()
    panel.processor.compute_enclosing_circle()
    panel.processor.compute_peak_detection(PeakDetectionParam())
    panel.processor.extract_roi(singleobj=singleobj)


def create_test_image_with_roi(size=None):
    """Create test image with ROIs"""
    ima = create_test_image3(size)
    dy, dx = ima.size
    roi1 = [dx // 2, dy // 2, dx - 25, dy]
    roi2 = [dx // 4, dy // 2, dx // 2, dy // 2]
    ima.roi = np.array([roi1, roi2], int)
    return ima


def array_2d_to_str(arr: np.ndarray) -> str:
    """Return 2-D array characteristics as string"""
    return f"{arr.shape[0]} x {arr.shape[1]} array (min={arr.min()}, max={arr.max()})"


def array_1d_to_str(arr: np.ndarray) -> str:
    """Return 1-D array characteristics as string"""
    return f"{arr.size} columns array (min={arr.min()}, max={arr.max()})"


def print_obj_shapes(obj):
    """Print object and associated ROI array shapes"""
    execenv.print(f"  Accessing object '{obj.title}':")
    func = array_1d_to_str if isinstance(obj, SignalParam) else array_2d_to_str
    execenv.print(f"    data: {func(obj.data)}")
    if obj.roi is not None:
        for idx in range(obj.roi.shape[0]):
            roi_data = obj.get_data(idx)
            if isinstance(obj, SignalParam):
                roi_data = roi_data[1]  # y data
            execenv.print(f"    ROI[{idx}]: {func(roi_data)}")


def test():
    """Run ROI unit test scenario"""
    size = 200
    with codraft_app_context() as win:
        execenv.print("ROI application test:")
        # === Signal ROI extraction test ===
        panel = win.signalpanel
        sig1 = create_test_signal1(size)
        panel.add_object(sig1)
        test_signal_features(panel)
        sig2 = create_test_signal1(size)
        sig2.roi = np.array([[26, 41], [125, 146]], int)
        for singleobj in (False, True):
            panel.add_object(sig2)
            print_obj_shapes(sig2)
            panel.processor.edit_regions_of_interest()
            win.take_screenshot("s_roi_signal")
            test_signal_features(panel, singleobj=singleobj)
        # === Image ROI extraction test ===
        panel = win.imagepanel
        ima1 = create_test_image3(size)
        panel.add_object(ima1)
        test_image_features(panel)
        ima2 = create_test_image_with_roi(size)
        for singleobj in (False, True):
            panel.add_object(ima2)
            print_obj_shapes(ima2)
            panel.processor.edit_regions_of_interest()
            win.take_screenshot("i_roi_image")
            test_image_features(panel, singleobj=singleobj)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT HDF5 open/save module
"""

import os.path as osp

from qtpy import QtWidgets as QW

from codraft.config import _
from codraft.core.io.base import NativeH5Reader, NativeH5Writer
from codraft.core.io.h5 import H5Importer
from codraft.core.model.signal import SignalParam
from codraft.utils.qthelpers import create_progress_bar, qt_try_loadsave_file
from codraft.widgets.h5browser import H5BrowserDialog


class H5InputOutput:
    """Object handling HDF5 file open/save into/from CodraFT data model/main window"""

    def __init__(self, mainwindow):
        self.mainwindow = mainwindow
        self.h5browser = None
        self.uint32_wng = None
        self.progressbar = None
        self.lmj_metadata = None

    @staticmethod
    def __progbartitle(fname):
        """Return progress bar title"""
        return _("Loading data from %s...") % osp.basename(fname)

    def save_file(self, filename):
        """Save all signals and images from CodraFT model into a HDF5 file"""
        writer = NativeH5Writer(filename)
        for panel in self.mainwindow.panels:
            panel.serialize_to_hdf5(writer)
        writer.close()

    def open_file(self, filename, import_all, reset_all):
        """Open HDF5 file"""
        progress = None
        try:
            reader = NativeH5Reader(filename)
            if reset_all:
                self.mainwindow.reset_all()
            with create_progress_bar(
                self.mainwindow, self.__progbartitle(filename), 2
            ) as progress:
                for idx, panel in enumerate(self.mainwindow.panels):
                    progress.setValue(idx)
                    QW.QApplication.processEvents()
                    panel.deserialize_from_hdf5(reader)
                    if progress.wasCanceled():
                        break
            reader.close()
        except KeyError:
            if progress is not None:
                # KeyError was encoutered when deserializing datasets (CodraFT data
                # model is not compatible with this version)
                progress.close()
            self.import_file(filename, import_all, reset_all)

    def __add_object_from_node(self, node):
        """Add CodraFT object from h5 node"""
        obj = node.get_object()
        self.uint32_wng = self.uint32_wng or node.uint32_wng
        if isinstance(obj, SignalParam):
            self.mainwindow.signalpanel.add_object(obj)
        else:
            self.mainwindow.imagepanel.add_object(obj)

    def __eventually_show_warnings(self):
        """Eventually show warnings after everything is imported"""
        if self.uint32_wng:
            QW.QMessageBox.warning(
                self.mainwindow, _("Warning"), _("Clipping uint32 data to int32.")
            )

    def import_file(self, filename, import_all, reset_all):
        """Import HDF5 file"""
        if self.h5browser is None:
            self.h5browser = H5BrowserDialog(self.mainwindow)

        with qt_try_loadsave_file(self.mainwindow, filename, "load"):
            self.h5browser.setup(filename)
            if not import_all and not self.h5browser.exec():
                self.h5browser.cleanup()
                return
            if import_all:
                nodes = self.h5browser.get_all_nodes()
            else:
                nodes = self.h5browser.get_nodes()
            if nodes is None:
                self.h5browser.cleanup()
                return
            if reset_all:
                self.mainwindow.reset_all()
            with create_progress_bar(
                self.mainwindow, self.__progbartitle(filename), len(nodes)
            ) as progress:
                self.uint32_wng = False
                for idx, node in enumerate(nodes):
                    progress.setValue(idx)
                    QW.QApplication.processEvents()
                    if progress.wasCanceled():
                        break
                    self.__add_object_from_node(node)
            self.h5browser.cleanup()
            self.__eventually_show_warnings()

    def import_dataset_from_file(self, filename, dsetname):
        """Import dataset from HDF5 file"""
        h5importer = H5Importer(filename)
        try:
            node = h5importer.get(dsetname)
            self.uint32_wng = False
            self.__add_object_from_node(node)
            self.__eventually_show_warnings()
        except KeyError as exc:
            raise KeyError(f"Dataset not found: {dsetname}") from exc
        h5importer.close()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT HDF5 browser module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import os
import os.path as osp

from guidata.qthelpers import (
    add_actions,
    create_action,
    create_toolbutton,
    get_icon,
    win32_fix_title_bar_background,
)
from guiqwt.plot import CurveWidget, ImageWidget
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from codraft.config import _
from codraft.core.io.h5 import H5Importer
from codraft.core.model.signal import SignalParam
from codraft.env import execenv
from codraft.utils.misc import to_string
from codraft.utils.qthelpers import qt_handle_error_message


class BaseTreeWidgetMeta(type(QW.QTreeWidget), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class BaseTreeWidget(QW.QTreeWidget, metaclass=BaseTreeWidgetMeta):
    """One-column tree widget with context menu, ..."""

    def __init__(self, parent):
        super().__init__(parent)
        self.setItemsExpandable(True)
        self.itemActivated.connect(self.activated)
        self.itemClicked.connect(self.clicked)
        # Setup context menu
        self.menu = QW.QMenu(self)
        self.collapse_all_action = None
        self.collapse_selection_action = None
        self.expand_all_action = None
        self.expand_selection_action = None
        self.common_actions = self.setup_common_actions()

        self.itemSelectionChanged.connect(self.item_selection_changed)
        self.item_selection_changed()

    @abc.abstractmethod
    def activated(self, item):
        """Double-click event"""

    @abc.abstractmethod
    def clicked(self, item):
        """Item was clicked"""

    def setup_common_actions(self):
        """Setup context menu common actions"""
        self.collapse_all_action = create_action(
            self,
            _("Collapse all"),
            icon=get_icon("collapse.png"),
            triggered=self.collapseAll,
        )
        self.expand_all_action = create_action(
            self, _("Expand all"), icon=get_icon("expand.png"), triggered=self.expandAll
        )
        self.restore_action = create_action(
            self,
            _("Restore"),
            tip=_("Restore original tree layout"),
            icon=get_icon("restore.png"),
            triggered=self.restore,
        )
        self.collapse_selection_action = create_action(
            self,
            _("Collapse selection"),
            icon=get_icon("collapse_selection.png"),
            triggered=self.collapse_selection,
        )
        self.expand_selection_action = create_action(
            self,
            _("Expand selection"),
            icon=get_icon("expand_selection.png"),
            triggered=self.expand_selection,
        )
        return [
            self.collapse_all_action,
            self.expand_all_action,
            self.restore_action,
            None,
            self.collapse_selection_action,
            self.expand_selection_action,
        ]

    def update_menu(self):
        """Update context menu"""
        self.menu.clear()
        items = self.selectedItems()
        actions = self.get_actions_from_items(items)
        if actions:
            actions.append(None)
        actions += self.common_actions
        add_actions(self.menu, actions)

    def get_actions_from_items(self, items):  # pylint: disable=W0613,R0201
        """Get actions from item"""
        # Right here: add other actions if necessary (reimplement this method)
        return []

    def restore(self):
        """Restore tree state"""
        self.collapseAll()
        for item in self.get_top_level_items():
            self.expandItem(item)

    def is_item_expandable(self, item):  # pylint: disable=W0613,R0201
        """To be reimplemented in child class
        See example in project explorer widget"""
        return True  # pragma: no cover

    def __expand_item(self, item):  # pragma: no cover
        """Expand item tree branch"""
        if self.is_item_expandable(item):
            self.expandItem(item)
            for index in range(item.childCount()):
                child = item.child(index)
                self.__expand_item(child)

    def expand_selection(self):  # pragma: no cover
        """Expand selection"""
        items = self.selectedItems()
        if not items:
            items = self.get_top_level_items()
        for item in items:
            self.__expand_item(item)
        if items:
            self.scrollToItem(items[0])

    def __collapse_item(self, item):  # pragma: no cover
        """Collapse item tree branch"""
        self.collapseItem(item)
        for index in range(item.childCount()):
            child = item.child(index)
            self.__collapse_item(child)

    def collapse_selection(self):  # pragma: no cover
        """Collapse selection"""
        items = self.selectedItems()
        if not items:
            items = self.get_top_level_items()
        for item in items:
            self.__collapse_item(item)
        if items:
            self.scrollToItem(items[0])

    def item_selection_changed(self):
        """Item selection has changed"""
        is_selection = len(self.selectedItems()) > 0
        self.expand_selection_action.setEnabled(is_selection)
        self.collapse_selection_action.setEnabled(is_selection)

    def get_top_level_items(self):
        """Iterate over top level items"""
        return [self.topLevelItem(_i) for _i in range(self.topLevelItemCount())]

    def get_items(self):
        """Return items (excluding top level items)"""
        itemlist = []

        def add_to_itemlist(item):
            for index in range(item.childCount()):
                citem = item.child(index)
                itemlist.append(citem)
                add_to_itemlist(citem)

        for tlitem in self.get_top_level_items():
            add_to_itemlist(tlitem)
        return itemlist

    def contextMenuEvent(self, event):
        """Override Qt method"""
        self.update_menu()
        self.menu.popup(event.globalPos())


class H5TreeWidget(BaseTreeWidget):
    """HDF5 Browser Tree Widget"""

    SIG_SELECTED = QC.Signal(QW.QTreeWidgetItem)

    def __init__(self, parent):
        super().__init__(parent)
        title = _("HDF5 Browser")
        self.setColumnCount(4)
        self.setWindowTitle(title)
        self.setHeaderLabels([_("Name"), _("Size"), _("Type"), _("Textual preview")])
        self.fname = None
        self.h5importer = None

    def setup(self, fname):
        """Setup H5TreeWidget"""
        self.fname = osp.abspath(fname)
        self.h5importer = H5Importer(self.fname)
        self.clear()
        self.populate_tree()
        self.expandAll()
        for col in range(3):
            self.resizeColumnToContents(col)

    def cleanup(self):
        """Clean up widget"""
        self.h5importer.close()
        self.h5importer = None

    def get_node(self, item):
        """Get HDF5 dataset associated to item"""
        node_id = item.data(0, QC.Qt.UserRole)
        if node_id:
            return self.h5importer.get(node_id)
        return None

    def get_nodes(self, only_checked_items=True):
        """Get all nodes associated to checked items"""
        datasets = []
        for item in self.find_all_items():
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                if only_checked_items and item.checkState(0) == 0:
                    continue
                if item is not self.topLevelItem(0):
                    node_id = item.data(0, QC.Qt.UserRole)
                    datasets.append(self.h5importer.get(node_id))
        return datasets

    def activated(self, item):
        """Double-click event"""
        if item is not self.topLevelItem(0):
            self.SIG_SELECTED.emit(item)

    def clicked(self, item):
        """Click event"""
        self.activated(item)

    def find_all_items(self):
        """Find all items"""
        return self.findItems("", QC.Qt.MatchContains | QC.Qt.MatchRecursive)

    def is_empty(self):
        """Return True if tree is empty"""
        return len(self.find_all_items()) == 1

    def get_item_from_id(self, item_id):
        """Return QListWidgetItem from id"""
        for item in self.find_all_items():
            if id(item) == item_id:
                return item
        return None

    def is_any_item_checked(self):
        """Return True if any item is checked"""
        for item in self.find_all_items():
            if item.checkState(0) > 0:
                return True
        return False

    def select_all(self, state):
        """Select all items"""
        for item in self.findItems("", QC.Qt.MatchContains | QC.Qt.MatchRecursive):
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                item.setSelected(state)
                if state:
                    self.clicked(item)

    def toggle_all(self, state):
        """Toggle all item state from 'unchecked' to 'checked'
        (or vice-versa)"""
        for item in self.findItems("", QC.Qt.MatchContains | QC.Qt.MatchRecursive):
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                item.setCheckState(0, QC.Qt.Checked if state else QC.Qt.Unchecked)

    @staticmethod
    def __create_node(node):
        text = to_string(node.text)
        if text:
            lines = text.splitlines()[:5]
            if len(lines) == 5:
                lines += ["[...]"]
            text = os.linesep.join(lines)
        treeitem = QW.QTreeWidgetItem([node.name, node.shape_str, node.dtype_str, text])
        treeitem.setData(0, QC.Qt.UserRole, node.id)
        if node.description:
            for col in range(treeitem.columnCount()):
                treeitem.setToolTip(col, node.description)
        return treeitem

    @staticmethod
    def __recursive_popfunc(parent_item, node):
        """Recursive HDF5 analysis"""
        tree_item = H5TreeWidget.__create_node(node)
        if node.IS_ARRAY:
            tree_item.setCheckState(0, QC.Qt.Unchecked)
        else:
            tree_item.setFlags(QC.Qt.ItemIsEnabled)
        tree_item.setIcon(0, get_icon(node.icon_name))
        parent_item.addChild(tree_item)
        for child in node.children:
            H5TreeWidget.__recursive_popfunc(tree_item, child)

    def populate_tree(self):
        """Populate tree"""
        root = self.h5importer.root
        rootitem = QW.QTreeWidgetItem([root.name])
        rootitem.setToolTip(0, root.description)
        rootitem.setData(0, QC.Qt.UserRole, root.id)
        rootitem.setFlags(QC.Qt.ItemIsEnabled)
        rootitem.setIcon(0, get_icon(root.icon_name))
        self.addTopLevelItem(rootitem)
        for node in root.children:
            self.__recursive_popfunc(rootitem, node)


class H5Browser(QW.QSplitter):
    """HDF5 Browser Widget"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.tree = H5TreeWidget(self)
        self.tree.SIG_SELECTED.connect(self.view_selected_item)
        self.addWidget(self.tree)
        self.stack = QW.QStackedWidget(self)
        self.addWidget(self.stack)
        self.curvewidget = CurveWidget(self.stack)
        self.curvewidget.register_all_curve_tools()
        self.stack.addWidget(self.curvewidget)
        self.imagewidget = ImageWidget(self.stack, show_contrast=True)
        self.imagewidget.register_all_image_tools()
        self.stack.addWidget(self.imagewidget)

    def setup(self, fname):
        """Setup widget"""
        self.tree.setup(fname)

    def cleanup(self):
        """Clean up widget"""
        for widget in (self.imagewidget, self.curvewidget):
            widget.plot.del_all_items()
        self.tree.cleanup()

    def get_node(self, item=None):
        """Return (selected) dataset"""
        if item is None:
            item = self.tree.currentItem()
        return self.tree.get_node(item)

    def view_selected_item(self, item):
        """View selected item"""
        node = self.get_node(item)
        if node.IS_ARRAY:
            self.update_visual_preview(node)

    def update_visual_preview(self, node):
        """Update visual preview widget"""
        try:
            obj = node.get_object()
        except Exception as msg:  # pylint: disable=broad-except
            qt_handle_error_message(self, msg)
            return
        if isinstance(obj, SignalParam):
            widget = self.curvewidget
        else:
            widget = self.imagewidget
        item = obj.make_item()
        plot = widget.plot
        plot.del_all_items()
        plot.add_item(item)
        plot.set_active_item(item)
        item.unselect()
        plot.do_autoscale()
        obj.update_item(item)
        self.stack.setCurrentWidget(widget)


class H5BrowserDialog(QW.QDialog):
    """HDF5 Browser Dialog"""

    def __init__(self, parent=None, size=(1550, 600)):
        super().__init__(parent)
        self.setWindowFlags(QC.Qt.Window)
        self.setObjectName("h5browser")
        self.setWindowTitle(_("HDF5 Browser"))
        self.setWindowIcon(get_icon("h5browser.svg"))
        win32_fix_title_bar_background(self)
        vlayout = QW.QVBoxLayout()
        self.setLayout(vlayout)
        self.button_layout = None
        self.bbox = None
        self.nodes = None

        self.browser = H5Browser(self)
        vlayout.addWidget(self.browser)

        self.browser.tree.itemChanged.connect(lambda item: self.refresh_buttons())

        self.install_button_layout()

        self.resize(QC.QSize(*size))
        self.browser.setSizes([int(size[1] / 2)] * 2)
        self.refresh_buttons()

    def accept(self):
        """Accept changes"""
        self.nodes = self.browser.tree.get_nodes()
        QW.QDialog.accept(self)

    def cleanup(self):
        """Cleanup dialog"""
        self.browser.cleanup()

    def refresh_buttons(self):
        """Refresh buttons"""
        state = self.browser.tree.is_any_item_checked()
        self.bbox.button(QW.QDialogButtonBox.Ok).setEnabled(state)

    def setup(self, fname):
        """Setup dialog"""
        self.browser.setup(fname)
        if self.browser.tree.is_empty():
            if not execenv.unattended:
                QW.QMessageBox.warning(
                    self.parent(),
                    self.windowTitle(),
                    _("Warning:")
                    + "\n"
                    + _("No supported data available in HDF5 file."),
                )
            QC.QTimer.singleShot(0, self.reject)

    def get_all_nodes(self):
        """Return all supported datasets"""
        return self.browser.tree.get_nodes(only_checked_items=False)

    def get_nodes(self):
        """Return datasets"""
        return self.nodes

    def install_button_layout(self):
        """Install button layout"""
        bbox = QW.QDialogButtonBox(QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel)
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)

        btn_check_all = create_toolbutton(
            self,
            text=_("Check all"),
            autoraise=False,
            shortcut=QG.QKeySequence.SelectAll,
            triggered=lambda checked=True: self.browser.tree.toggle_all(checked),
        )
        btn_uncheck_all = create_toolbutton(
            self,
            text=_("Uncheck all"),
            autoraise=False,
            triggered=lambda checked=False: self.browser.tree.toggle_all(checked),
        )

        self.button_layout = QW.QHBoxLayout()
        self.button_layout.addWidget(btn_check_all)
        self.button_layout.addWidget(btn_uncheck_all)
        self.button_layout.addStretch()
        self.button_layout.addWidget(bbox)
        self.bbox = bbox

        vlayout = self.layout()
        vlayout.addSpacing(10)
        vlayout.addLayout(self.button_layout)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT, the Codra Filtering Tool
Simple signal and image processing application based on guiqwt and guidata

Starter
"""

from codraft.app import run

run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""Curve fitting dialog test

Testing fit dialogs: Gaussian, Lorentzian, Voigt, etc.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from codraft.core.computation.signal import peak_indexes
from codraft.core.io.signal import read_signal
from codraft.env import execenv
from codraft.tests.data import create_1d_gaussian, get_test_fnames
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.tests import get_default_test_name
from codraft.widgets import fitdialog as fdlg

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Test function"""
    with qt_app_context():

        # Multi-gaussian curve fitting test
        s = read_signal(get_test_fnames("paracetamol.txt")[0])
        peakidx = peak_indexes(s.y)
        execenv.print(
            fdlg.multigaussianfit(s.x, s.y, peakidx, name=get_default_test_name("00"))
        )

        # Gaussian curve fitting test
        size = 500
        x, y = create_1d_gaussian(size=size, noise_sigma=5.0)
        execenv.print(fdlg.gaussianfit(x, y))

        # Lorentzian curve fitting test
        execenv.print(fdlg.lorentzianfit(x, y))

        # Voigt curve fitting test
        execenv.print(fdlg.voigtfit(x, y))

        # Polynomial curve fitting test
        execenv.print(fdlg.polynomialfit(x, y, 4))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image peak detection test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

import time

from guiqwt.builder import make

from codraft.core.computation.image import get_2d_peaks_coords
from codraft.env import execenv
from codraft.tests.data import get_peak2d_data
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_image_items

SHOW = True  # Show test in GUI-based test launcher


def exec_image_peak_detection_func(data):
    """Compare image peak detection methods"""
    items = [make.image(data, interpolation="linear", colormap="hsv")]
    t0 = time.time()
    coords = get_2d_peaks_coords(data)
    dt = time.time() - t0
    for x, y in coords:
        items.append(make.marker((x, y)))
    execenv.print(f"Calculation time: {int(dt * 1e3):d} ms")
    execenv.print(f"  => {coords.tolist()}")
    view_image_items(items)


def peak2d_test():
    """2D peak detection test"""
    with qt_app_context():
        exec_image_peak_detection_func(get_peak2d_data(multi=False))


if __name__ == "__main__":
    peak2d_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Metadata import/export unit test:

  - Create an image with annotations and result shapes
  - Add the image to CodraFT
  - Export image metadata to file (JSON)
  - Delete image metadata
  - Import image metadata from previous file
  - Check if image metadata is the same as the original image
"""

import os.path as osp

import numpy as np

from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.tests import data as test_data
from codraft.utils import tests

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run image tools test scenario"""
    execenv.unattended = True
    with tests.temporary_directory() as tmpdir:
        fname = osp.join(tmpdir, "test.json")
        with codraft_app_context() as win:
            panel = win.imagepanel
            ima = test_data.create_image_with_annotations()
            for mshape in test_data.create_resultshapes():
                mshape.add_to(ima)
            panel.add_object(ima)
            orig_metadata = ima.metadata.copy()
            panel.export_metadata_from_file(fname)
            panel.delete_metadata()
            assert len(ima.metadata) == 0
            panel.import_metadata_from_file(fname)
            execenv.print("Check metadata export <--> import features:")
            for key, value in orig_metadata.items():
                execenv.print(f"  Checking {key} key value...", end="")
                if isinstance(value, np.ndarray):
                    assert (value == ima.metadata[key]).all()
                else:
                    assert value == ima.metadata[key]
                execenv.print("OK")


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Base I/O common module (native HDF5 format)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from guidata.hdf5io import HDF5Reader, HDF5Writer

from codraft import __version__

H5_VERSION = "CodraFT_Version"


LIST_LENGTH_STR = "__list_length__"


class NativeH5Writer(HDF5Writer):
    """CodraFT signal/image objects HDF5 guidata Dataset Writer class,
    supporting dictionary serialization"""

    def __init__(self, filename):
        super().__init__(filename)
        self.h5[H5_VERSION] = __version__

    def write_dict(self, val):
        """Write dictionary to h5 file"""
        # Keys must be strings
        # Values must be h5py supported data types
        group = self.get_parent_group()
        dict_group = group.create_group(self.option[-1])
        for key, value in val.items():
            if isinstance(value, dict):
                with self.group(key):
                    self.write_dict(value)
            elif isinstance(value, list):
                with self.group(key):
                    with self.group(LIST_LENGTH_STR):
                        self.write(len(value))
                    for index, i_val in enumerate(value):
                        with self.group("elt" + str(index)):
                            self.write(i_val)
            else:
                try:
                    dict_group.attrs[key] = value
                except TypeError:
                    pass


class NativeH5Reader(HDF5Reader):
    """CodraFT signal/image objects HDF5 guidata dataset Writer class,
    supporting dictionary deserialization"""

    def __init__(self, filename):
        super().__init__(filename)
        self.version = self.h5[H5_VERSION]

    def read_dict(self):
        """Read dictionary from h5 file"""
        group = self.get_parent_group()
        dict_group = group[self.option[-1]]
        dict_val = {}
        for key, value in dict_group.attrs.items():
            dict_val[key] = value
        for key in dict_group:
            with self.group(key):
                if "__list_length__" in dict_group[key].attrs:
                    with self.group(LIST_LENGTH_STR):
                        list_len = self.read()
                    dict_val[key] = [
                        dict_group[key]["elt" + str(index)][:]
                        for index in range(list_len)
                    ]
                else:
                    dict_val[key] = self.read_dict()
        return dict_val



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Metadata application test:

  - Create signal/image, with ROI
  - Compute things (adds metadata)
  - Test metadata delete, copy, paste
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np

from codraft.core.gui.panel import BasePanel, ImagePanel, SignalPanel
from codraft.core.gui.processor.image import PeakDetectionParam
from codraft.core.gui.processor.signal import FWHMParam
from codraft.env import execenv
from codraft.tests import codraft_app_context, roi_app
from codraft.tests.data import create_test_signal1

SHOW = True  # Show test in GUI-based test launcher


def test_signal_features(panel: SignalPanel):
    """Test all signal features related to ROI"""
    panel.processor.compute_fwhm(FWHMParam())
    panel.processor.compute_fw1e2()


def test_image_features(panel: ImagePanel):
    """Test all image features related to ROI"""
    panel.processor.compute_centroid()
    panel.processor.compute_enclosing_circle()
    panel.processor.compute_peak_detection(PeakDetectionParam())


def test_metadata_features(panel: BasePanel):
    """Test all metadata features"""
    panel.duplicate_object()
    panel.delete_metadata()
    panel.objlist.set_current_row(-2)
    panel.copy_metadata()
    panel.objlist.set_current_row(-1)
    panel.paste_metadata()


def test():
    """Run ROI unit test scenario"""
    size = 200
    with codraft_app_context() as win:
        execenv.print("Metadata application test:")
        # === Signal metadata features test ===
        panel = win.signalpanel
        sig = create_test_signal1(size)
        sig.roi = np.array([[26, 41], [125, 146]], int)
        panel.add_object(sig)
        test_signal_features(panel)
        test_metadata_features(panel)
        # === Image metadata features test ===
        panel = win.imagepanel
        ima = roi_app.create_test_image_with_roi(size)
        panel.add_object(ima)
        test_image_features(panel)
        test_metadata_features(panel)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Module patching *guiqwt* to adapt it to CodraFT
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# Allows accessing protecting members, unused arguments, unused variables
# pylint: disable=W0212,W0613,W0612,E0102

import sys
import warnings

import guidata.dataset.datatypes
import guiqwt.annotations
import guiqwt.curve
import guiqwt.histogram
import guiqwt.image
import guiqwt.plot
import guiqwt.tools
import numpy as np
from guidata.configtools import get_icon
from guidata.dataset.qtwidgets import DataSetEditDialog, DataSetGroupEditDialog
from guidata.qthelpers import add_actions, create_action
from guiqwt import cross_section as cs
from guiqwt.transitional import QwtLinearScaleEngine
from qtpy.QtWidgets import QApplication, QMainWindow
from qwt import QwtLogScaleEngine as QwtLog10ScaleEngine
from qwt import QwtScaleDraw

from codraft.config import APP_NAME, _
from codraft.core.model.signal import create_signal
from codraft.utils.qthelpers import block_signals, exec_dialog


def monkeypatch_method(cls, patch_name):
    # This function's code was inspired from the following thread:
    # "[Python-Dev] Monkeypatching idioms -- elegant or ugly?"
    # by Robert Brewer <fumanchu at aminus.org>
    # (Tue Jan 15 19:13:25 CET 2008)
    """
    Add the decorated method to the given class; replace as needed.

    If the named method already exists on the given class, it will
    be replaced, and a reference to the old method is created as
    cls._old<patch_name><name>. If the "_old_<patch_name>_<name>" attribute
    already exists, KeyError is raised.
    """

    def decorator(func):
        """Decorateur wrapper function"""
        fname = func.__name__
        old_func = getattr(cls, fname, None)
        if old_func is not None:
            # Add the old func to a list of old funcs.
            old_ref = f"_old_{patch_name}_{fname}"
            # print old_ref, old_func
            old_attr = getattr(cls, old_ref, None)
            if old_attr is None:
                setattr(cls, old_ref, old_func)
            else:
                print(
                    f"Warning: {cls.__name__}.{fname} already patched",
                    file=sys.stderr,
                )
        setattr(cls, fname, func)
        return func

    return decorator


# TODO: Improve test coverage using the following two patches
# (sometimes we had to skip DataSet edit method to avoid blocking GUI testing:
# now it's possible to test those lines without blocking)

#  Patching guidata.dataset.datatypes.DataSet edit methods for automatic GUI testing
@monkeypatch_method(guidata.dataset.datatypes.DataSet, "DataSet")
def edit(self, parent=None, apply=None, size=None):
    """
    Open a dialog box to edit data set
        * parent: parent widget (default is None, meaning no parent)
        * apply: apply callback (default is None)
        * size: dialog size (QSize object or integer tuple (width, height))
    """
    icon = self.get_icon()
    win = DataSetEditDialog(self, icon=icon, parent=parent, apply=apply, size=size)
    return exec_dialog(win)


#  Patching guidata.dataset.datatypes.DataSet edit methods for automatic GUI testing
@monkeypatch_method(guidata.dataset.datatypes.DataSetGroup, "DataSetGroup")
def edit(self, parent=None, apply=None, size=None):
    """
    Open a dialog box to edit data set
    """
    win = DataSetGroupEditDialog(self, parent=parent, apply=apply)
    return exec_dialog(win)


# Patching AnnotatedSegment "get_infos" method for a more compact text
@monkeypatch_method(guiqwt.annotations.AnnotatedSegment, "AnnotatedSegment")
def get_infos(self):
    """Return formatted string with informations on current shape"""
    return "Δ = " + self.x_to_str(self.get_tr_length())


#  Patching CurveItem's "select" method to avoid showing giant ugly squares
@monkeypatch_method(guiqwt.curve.CurveItem, "CurveItem")
def select(self):
    """Select item"""
    self.selected = True
    plot = self.plot()
    with block_signals(widget=plot, enable=plot is not None):
        pen = self.curveparam.line.build_pen()
        pen.setWidth(2)
        self.setPen(pen)
    self.invalidate_plot()


#  Adding centroid parameter to the image stats tool
@monkeypatch_method(guiqwt.image.BaseImageItem, "ImageItem")
def get_stats(self, x0, y0, x1, y1):
    """Return formatted string with stats on image rectangular area
    (output should be compatible with AnnotatedShape.get_infos)"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        txt = self._old_ImageItem_get_stats(x0, y0, x1, y1)
    ix0, iy0, ix1, iy1 = self.get_closest_index_rect(x0, y0, x1, y1)
    data = self.data[iy0:iy1, ix0:ix1]

    # pylint: disable=C0415
    from codraft.core.computation.image import get_centroid_fourier

    c_i, c_j = get_centroid_fourier(data)
    c_x, c_y = self.get_plot_coordinates(c_j + ix0, c_i + iy0)
    xfmt = self.imageparam.xformat
    yfmt = self.imageparam.yformat
    return (
        txt
        + "<br>"
        + "<br>".join(
            [
                "C|x = " + xfmt % c_x,
                "C|y = " + yfmt % c_y,
            ]
        )
    )


# ==============================================================================
#  Adding support for z-axis logarithmic scale
# ==============================================================================
class ZAxisLogTool(guiqwt.tools.ToggleTool):
    """Patched tools.ToggleTool"""

    def __init__(self, manager):
        title = _("Z-Axis logarithmic scale")
        super().__init__(
            manager,
            title=title,
            toolbar_id=guiqwt.tools.DefaultToolbarID,
            icon="zlog.svg",
        )

    def activate_command(self, plot, checked):
        """Reimplement tools.ToggleTool method"""
        for item in self.get_supported_items(plot):
            item.set_zaxis_log_state(not item.get_zaxis_log_state())
        plot.replot()
        self.update_status(plot)

    def get_supported_items(self, plot):
        """Reimplement tools.ToggleTool method"""
        items = [
            item
            for item in plot.get_items()
            if isinstance(item, guiqwt.image.ImageItem)
            and not item.is_empty()
            and hasattr(item, "get_zaxis_log_state")
        ]
        if len(items) > 1:
            items = [item for item in items if item in plot.get_selected_items()]
        if items:
            self.action.setChecked(items[0].get_zaxis_log_state())
        return items

    def update_status(self, plot):
        """Reimplement tools.ToggleTool method"""
        self.action.setEnabled(len(self.get_supported_items(plot)) > 0)


@monkeypatch_method(guiqwt.plot.PlotManager, "PlotManager")
def register_image_tools(self):
    """Reimplement guiqwt.plot.PlotManager method"""
    self._old_PlotManager_register_image_tools()
    self.add_tool(ZAxisLogTool)


@monkeypatch_method(guiqwt.image.ImageItem, "ImageItem")
def __init__(self, data=None, param=None):
    self._log_data = None
    self._lin_lut_range = None
    self._is_zaxis_log = False
    self._old_ImageItem___init__(data=data, param=param)


class ZLogScaleDraw(QwtScaleDraw):
    """Patched QwtScaleDraw"""

    def label(self, value):
        """Reimplement QwtScaleDraw method"""
        logvalue = int(10 ** min([100.0, value]) - 1)
        return super().label(logvalue)
        # XXX: [P3] This is not the cleanest way to show log-Z scale
        # (we should be able to trick PyQwt at another stage of the
        #  rendering process, in order to choose scale data properly)


@monkeypatch_method(guiqwt.image.ImageItem, "ImageItem")
def set_zaxis_log_state(self, state):
    """Reimplement image.ImageItem method"""
    self._is_zaxis_log = state
    plot = self.plot()
    if state:
        self._lin_lut_range = self.get_lut_range()
        if self._log_data is None:
            self._log_data = np.array(
                np.log10(self.data.clip(1)), dtype=np.float64, copy=True
            )
        self.set_lut_range(
            [guiqwt.image._nanmin(self._log_data), guiqwt.image._nanmax(self._log_data)]
        )
        plot.setAxisScaleDraw(plot.yRight, ZLogScaleDraw())
        plot.setAxisScaleEngine(plot.yRight, QwtLog10ScaleEngine())
    else:
        self._log_data = None
        self.set_lut_range(self._lin_lut_range)
        plot.setAxisScaleDraw(plot.yRight, QwtScaleDraw())
        plot.setAxisScaleEngine(plot.yRight, QwtLinearScaleEngine())
    plot.update_colormap_axis(self)


@monkeypatch_method(guiqwt.image.ImageItem, "ImageItem")
def get_zaxis_log_state(self):
    """Reimplement image.ImageItem method"""
    return self._is_zaxis_log


@monkeypatch_method(guiqwt.image.ImageItem, "ImageItem")
def draw_image(self, painter, canvasRect, src_rect, dst_rect, xMap, yMap):
    """Reimplement image.ImageItem method"""
    if self.data is None:
        return
    src2 = self._rescale_src_rect(src_rect)
    dst_rect = tuple(int(i) for i in dst_rect)

    # Not the most efficient way to do it, but it works...
    # --------------------------------------------------------------------------
    if self.get_zaxis_log_state():
        data = self._log_data
    else:
        data = self.data
    # --------------------------------------------------------------------------

    dest = guiqwt.image._scale_rect(
        data, src2, self._offscreen, dst_rect, self.lut, (guiqwt.image.INTERP_NEAREST,)
    )
    qrect = guiqwt.image.QRectF(
        guiqwt.image.QPointF(dest[0], dest[1]), guiqwt.image.QPointF(dest[2], dest[3])
    )
    painter.drawImage(qrect, self._image, qrect)


# ==============================================================================
#  Cross section : add a button to send curve to CodraFT's signal panel
# ==============================================================================
def to_codraft(cs_plot):
    """Send cross section curve to CodraFT's signal list"""
    win = None
    for win in QApplication.topLevelWidgets():
        if isinstance(win, QMainWindow):
            break
    if win is None or win.objectName() != APP_NAME:
        # pylint: disable=import-outside-toplevel
        # pylint: disable=cyclic-import
        from codraft.core.gui import main

        # Note : this is the only way to retrieve the CodraFT main window instance
        # when the CrossSectionItem object is embedded into an image widget
        # parented to another main window.
        win = main.CodraFTMainWindow.get_instance()
        assert win is not None  # Should never happen

    for item in cs_plot.get_items():
        if not isinstance(item, guiqwt.curve.CurveItem):
            continue
        x, y, _dx, _dy = item.get_data()
        if x is None or y is None or x.size == 0 or y.size == 0:
            continue

        signal = create_signal(item.curveparam.label)

        image_item = None
        for image_item, curve_item in cs_plot.known_items.items():
            if curve_item is item:
                break
        image_plot = image_item.plot()

        if isinstance(cs_plot, cs.VerticalCrossSectionPlot):
            signal.set_xydata(y, x)
            xaxis_name = "left"
            xunit = image_plot.get_axis_unit("bottom")
            if xunit:
                signal.title += " " + xunit
        else:
            signal.set_xydata(x, y)
            xaxis_name = "bottom"
            yunit = image_plot.get_axis_unit("left")
            if yunit:
                signal.title += " " + yunit

        signal.ylabel = image_plot.get_axis_title("right")
        signal.yunit = image_plot.get_axis_unit("right")
        signal.xlabel = image_plot.get_axis_title(xaxis_name)
        signal.xunit = image_plot.get_axis_unit(xaxis_name)

        win.signalpanel.add_object(signal)

    # Show CodraFT main window on top, if not already visible
    win.show()
    win.raise_()


@monkeypatch_method(cs.XCrossSection, "XCrossSection")
def add_actions_to_toolbar(self):
    """Add actions to toolbar"""
    to_codraft_ac = create_action(
        self,
        _("Process signal"),
        icon=get_icon("to_signal.svg"),
        triggered=lambda: to_codraft(self.cs_plot),
    )
    add_actions(self.toolbar, (to_codraft_ac, None))
    self._old_XCrossSection_add_actions_to_toolbar()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Action handler module

This module handles all application actions (menus, toolbars, context menu).
These actions point to CodraFT panels, processors, objectlist, ...
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import enum

from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from codraft.config import _
from codraft.widgets import fitdialog


class ActionCategory(enum.Enum):
    """Action categories"""

    FILE = enum.auto()
    EDIT = enum.auto()
    VIEW = enum.auto()
    OPERATION = enum.auto()
    PROCESSING = enum.auto()
    COMPUTING = enum.auto()


class BaseActionHandler(metaclass=abc.ABCMeta):
    """Object handling panel GUI interactions: actions, menus, ..."""

    OBJECT_STR = ""  # e.g. "signal"

    def __init__(self, panel, objlist, itmlist, processor, toolbar):
        self.panel = panel
        self.objlist = objlist
        self.itmlist = itmlist
        self.processor = processor
        self.feature_actions = {}
        self.operation_end_actions = None
        self.delete_roi_action = None
        # Object selection dependent actions
        self.actlist_1more = []
        self.actlist_2more = []
        self.actlist_1 = []
        self.actlist_2 = []
        self.actlist_cmenu = []  # Context menu
        if self.__class__ is not BaseActionHandler:
            self.create_all_actions(toolbar)

    def get_context_menu_actions(self):
        """Return context menu action list"""
        return self.actlist_cmenu

    def selection_rows_changed(self):
        """Number of selected rows has changed"""
        selrows = self.objlist.get_selected_rows()
        nbrows = len(selrows)
        for act in self.actlist_1more:
            act.setEnabled(nbrows >= 1)
        for act in self.actlist_2more:
            act.setEnabled(nbrows >= 2)
        for act in self.actlist_1:
            act.setEnabled(nbrows == 1)
        for act in self.actlist_2:
            act.setEnabled(nbrows == 2)
        self.delete_roi_action.setEnabled(False)
        for row in selrows:
            obj = self.objlist[row]
            if obj.roi is not None:
                self.delete_roi_action.setEnabled(True)
                break

    def create_all_actions(self, toolbar):
        """Setup actions, menus, toolbar"""
        featact = self.feature_actions
        featact[ActionCategory.FILE] = file_act = self.create_file_actions()
        featact[ActionCategory.EDIT] = edit_act = self.create_edit_actions()
        featact[ActionCategory.VIEW] = view_act = self.create_view_actions()
        featact[ActionCategory.OPERATION] = self.create_operation_actions()
        featact[ActionCategory.PROCESSING] = self.create_processing_actions()
        featact[ActionCategory.COMPUTING] = self.create_computing_actions()
        add_actions(toolbar, file_act + [None] + edit_act + [None] + view_act)

    def cra(
        self, title, triggered=None, toggled=None, shortcut=None, icon=None, tip=None
    ):
        """Create action convenience method"""
        return create_action(self.panel, title, triggered, toggled, shortcut, icon, tip)

    def create_file_actions(self):
        """Create file actions"""
        new_act = self.cra(
            _("New %s...") % self.OBJECT_STR,
            icon=get_icon(f"new_{self.OBJECT_STR}.svg"),
            tip=_("Create new %s") % self.OBJECT_STR,
            triggered=self.panel.new_object,
            shortcut=QG.QKeySequence(QG.QKeySequence.New),
        )
        open_act = self.cra(
            _("Open %s...") % self.OBJECT_STR,
            icon=get_icon("libre-gui-import.svg"),
            tip=_("Open %s") % self.OBJECT_STR,
            triggered=self.panel.open_objects,
            shortcut=QG.QKeySequence(QG.QKeySequence.Open),
        )
        save_act = self.cra(
            _("Save %s...") % self.OBJECT_STR,
            icon=get_icon("libre-gui-export.svg"),
            tip=_("Save selected %s") % self.OBJECT_STR,
            triggered=self.panel.save_objects,
            shortcut=QG.QKeySequence(QG.QKeySequence.Save),
        )
        importmd_act = self.cra(
            _("Import metadata into %s...") % self.OBJECT_STR,
            icon=get_icon("metadata_import.svg"),
            tip=_("Import metadata into %s") % self.OBJECT_STR,
            triggered=self.panel.import_metadata_from_file,
        )
        exportmd_act = self.cra(
            _("Export metadata from %s...") % self.OBJECT_STR,
            icon=get_icon("metadata_export.svg"),
            tip=_("Export selected %s metadata") % self.OBJECT_STR,
            triggered=self.panel.export_metadata_from_file,
        )
        self.actlist_1more += [save_act]
        self.actlist_cmenu += [save_act]
        self.actlist_1 += [importmd_act, exportmd_act]
        return [new_act, open_act, save_act, None, importmd_act, exportmd_act]

    def create_edit_actions(self):
        """Create edit actions"""
        dup_action = self.cra(
            _("Duplicate"),
            icon=get_icon("libre-gui-copy.svg"),
            triggered=self.panel.duplicate_object,
            shortcut=QG.QKeySequence(QG.QKeySequence.Copy),
        )
        cpymeta_action = self.cra(
            _("Copy metadata"),
            icon=get_icon("metadata_copy.svg"),
            triggered=self.panel.copy_metadata,
        )
        pstmeta_action = self.cra(
            _("Paste metadata"),
            icon=get_icon("metadata_paste.svg"),
            triggered=self.panel.paste_metadata,
        )
        cleanup_action = self.cra(
            _("Clean up data view"),
            icon=get_icon("libre-tools-vacuum-cleaner.svg"),
            tip=_("Clean up data view before updating plotting panels"),
            toggled=self.itmlist.toggle_cleanup_dataview,
        )
        cleanup_action.setChecked(True)
        delm_action = self.cra(
            _("Delete object metadata"),
            icon=get_icon("metadata_delete.svg"),
            tip=_("Delete all that is contained in object metadata"),
            triggered=self.panel.delete_metadata,
        )
        delall_action = self.cra(
            _("Delete all"),
            shortcut="Shift+Ctrl+Suppr",
            icon=get_icon("delete_all.svg"),
            triggered=self.panel.delete_all_objects,
        )
        del_action = self.cra(
            _("Remove"),
            icon=get_icon("delete.svg"),
            triggered=self.panel.remove_object,
            shortcut=QG.QKeySequence(QG.QKeySequence.Delete),
        )
        self.actlist_1more += [
            dup_action,
            del_action,
            delm_action,
            pstmeta_action,
            delall_action,
        ]
        self.actlist_cmenu += [dup_action, del_action]
        self.actlist_1 += [cpymeta_action]
        return [
            dup_action,
            del_action,
            delall_action,
            None,
            cpymeta_action,
            pstmeta_action,
            delm_action,
        ]

    def create_view_actions(self):
        """Create view actions"""
        view_action = self.cra(
            _("View in a new window"),
            icon=get_icon("libre-gui-binoculars.svg"),
            triggered=self.panel.open_separate_view,
        )
        showlabel_action = self.cra(
            _("Show graphical object titles"),
            icon=get_icon("show_titles.svg"),
            tip=_("Show or hide ROI and other graphical object titles or subtitles"),
            toggled=self.panel.toggle_show_titles,
        )
        showlabel_action.setChecked(False)
        self.actlist_1more += [view_action]
        self.actlist_cmenu = [view_action, None] + self.actlist_cmenu
        return [view_action, showlabel_action]

    def create_operation_actions(self):
        """Create operation actions"""
        proc = self.processor
        sum_action = self.cra(_("Sum"), proc.compute_sum)
        average_action = self.cra(_("Average"), proc.compute_average)
        diff_action = self.cra(_("Difference"), lambda: proc.compute_difference(False))
        qdiff_action = self.cra(
            _("Quadratic difference"), lambda: proc.compute_difference(True)
        )
        prod_action = self.cra(_("Product"), proc.compute_product)
        div_action = self.cra(_("Division"), proc.compute_division)
        roi_action = self.cra(
            _("ROI extraction"),
            proc.extract_roi,
            icon=get_icon(f"{self.OBJECT_STR}_roi.svg"),
        )
        swapaxes_action = self.cra(_("Swap X/Y axes"), proc.swap_axes)
        abs_action = self.cra(_("Absolute value"), proc.compute_abs)
        log_action = self.cra("Log10(y)", proc.compute_log10)
        self.actlist_1more += [roi_action, swapaxes_action, abs_action, log_action]
        self.actlist_2more += [sum_action, average_action, prod_action]
        self.actlist_2 += [diff_action, qdiff_action, div_action]
        self.operation_end_actions = [roi_action, swapaxes_action]
        return [
            sum_action,
            average_action,
            diff_action,
            qdiff_action,
            prod_action,
            div_action,
            None,
            abs_action,
            log_action,
        ]

    def create_processing_actions(self):
        """Create processing actions"""
        proc = self.processor
        threshold_action = self.cra(_("Thresholding"), proc.compute_threshold)
        clip_action = self.cra(_("Clipping"), proc.compute_clip)
        lincal_action = self.cra(_("Linear calibration"), proc.calibrate)
        gauss_action = self.cra(_("Gaussian filter"), proc.compute_gaussian)
        movavg_action = self.cra(_("Moving average"), proc.compute_moving_average)
        movmed_action = self.cra(_("Moving median"), proc.compute_moving_median)
        wiener_action = self.cra(_("Wiener filter"), proc.compute_wiener)
        fft_action = self.cra(_("FFT"), proc.compute_fft)
        ifft_action = self.cra(_("Inverse FFT"), proc.compute_ifft)
        for act in (fft_action, ifft_action):
            act.setToolTip(_("Warning: only real part is plotted"))
        actions = [
            threshold_action,
            clip_action,
            lincal_action,
            gauss_action,
            movavg_action,
            movmed_action,
            wiener_action,
            fft_action,
            ifft_action,
        ]
        self.actlist_1more += actions
        return actions

    @abc.abstractmethod
    def create_computing_actions(self):
        """Create computing actions"""
        proc = self.processor
        defineroi_action = self.cra(
            _("Edit regions of interest..."),
            triggered=proc.edit_regions_of_interest,
            icon=get_icon("roi.svg"),
        )
        self.delete_roi_action = self.cra(
            _("Remove regions of interest"),
            triggered=proc.delete_regions_of_interest,
            icon=get_icon("roi_delete.svg"),
        )
        stats_action = self.cra(
            _("Statistics") + "...",
            triggered=proc.compute_stats,
            icon=get_icon("stats.svg"),
        )
        self.actlist_1 += [defineroi_action, self.delete_roi_action, stats_action]
        self.actlist_cmenu += [
            None,
            defineroi_action,
            self.delete_roi_action,
            None,
            stats_action,
        ]
        return [defineroi_action, self.delete_roi_action, None, stats_action]


class SignalActionHandler(BaseActionHandler):
    """Object handling signal panel GUI interactions: actions, menus, ..."""

    OBJECT_STR = _("signal")

    def create_operation_actions(self):
        """Create operation actions"""
        base_actions = super().create_operation_actions()
        proc = self.processor
        peakdetect_action = self.cra(
            _("Peak detection"),
            proc.detect_peaks,
            icon=get_icon("peak_detect.svg"),
        )
        self.actlist_1more += [peakdetect_action]
        roi_actions = self.operation_end_actions
        return base_actions + [None, peakdetect_action, None] + roi_actions

    def create_processing_actions(self):
        """Create processing actions"""
        base_actions = super().create_processing_actions()
        proc = self.processor
        normalize_action = self.cra(_("Normalize"), proc.normalize)
        deriv_action = self.cra(_("Derivative"), proc.compute_derivative)
        integ_action = self.cra(_("Integral"), proc.compute_integral)
        polyfit_action = self.cra(_("Polynomial fit"), proc.compute_polyfit)
        mgfit_action = self.cra(_("Multi-Gaussian fit"), proc.compute_multigaussianfit)

        def cra_fit(title, fitdlgfunc):
            """Create curve fitting action"""
            return self.cra(title, lambda: proc.compute_fit(title, fitdlgfunc))

        gaussfit_action = cra_fit(_("Gaussian fit"), fitdialog.gaussianfit)
        lorentzfit_action = cra_fit(_("Lorentzian fit"), fitdialog.lorentzianfit)
        voigtfit_action = cra_fit(_("Voigt fit"), fitdialog.voigtfit)
        actions1 = [normalize_action, deriv_action, integ_action]
        actions2 = [
            gaussfit_action,
            lorentzfit_action,
            voigtfit_action,
            polyfit_action,
            mgfit_action,
        ]
        self.actlist_1more += actions1 + actions2
        return actions1 + [None] + base_actions + [None] + actions2

    def create_computing_actions(self):
        """Create computing actions"""
        base_actions = super().create_computing_actions()
        proc = self.processor
        fwhm_action = self.cra(
            _("Full width at half-maximum"),
            triggered=proc.compute_fwhm,
            tip=_("Compute Full Width at Half-Maximum (FWHM)"),
        )
        fw1e2_action = self.cra(
            _("Full width at") + " 1/e²",
            triggered=proc.compute_fw1e2,
            tip=_("Compute Full Width at Maximum") + "/e²",
        )
        self.actlist_1more += [fwhm_action, fw1e2_action]
        return base_actions + [fwhm_action, fw1e2_action]


class ImageActionHandler(BaseActionHandler):
    """Object handling image panel GUI interactions: actions, menus, ..."""

    OBJECT_STR = _("image")

    def create_view_actions(self):
        """Create view actions"""
        base_actions = super().create_view_actions()
        showcontrast_action = self.cra(
            _("Show contrast panel"),
            icon=get_icon("contrast.png"),
            tip=_("Show or hide contrast adjustment panel"),
            toggled=self.panel.toggle_show_contrast,
        )
        showcontrast_action.setChecked(True)
        self.actlist_1more += [showcontrast_action]
        return base_actions + [showcontrast_action]

    def create_operation_actions(self):
        """Create operation actions"""
        base_actions = super().create_operation_actions()
        proc = self.processor
        rotate_menu = QW.QMenu(_("Rotation"), self.panel)
        hflip_act = self.cra(
            _("Flip horizontally"),
            triggered=proc.flip_horizontally,
            icon=get_icon("flip_horizontally.svg"),
        )
        vflip_act = self.cra(
            _("Flip vertically"),
            triggered=proc.flip_vertically,
            icon=get_icon("flip_vertically.svg"),
        )
        rot90_act = self.cra(
            _("Rotate %s right") % "90°",  # pylint: disable=consider-using-f-string
            triggered=proc.rotate_270,
            icon=get_icon("rotate_right.svg"),
        )
        rot270_act = self.cra(
            _("Rotate %s left") % "90°",  # pylint: disable=consider-using-f-string
            triggered=proc.rotate_90,
            icon=get_icon("rotate_left.svg"),
        )
        rotate_act = self.cra(
            _("Rotate arbitrarily..."), triggered=proc.rotate_arbitrarily
        )
        resize_act = self.cra(_("Resize"), triggered=proc.resize_image)
        logp1_act = self.cra("Log10(z+n)", triggered=proc.compute_logp1)
        flatfield_act = self.cra(
            _("Flat-field correction"), triggered=proc.flat_field_correction
        )
        self.actlist_2 += [flatfield_act]
        self.actlist_1more += [
            resize_act,
            hflip_act,
            vflip_act,
            logp1_act,
            rot90_act,
            rot270_act,
            rotate_act,
        ]
        self.actlist_cmenu += [None, hflip_act, vflip_act, rot90_act, rot270_act]
        add_actions(
            rotate_menu, [hflip_act, vflip_act, rot90_act, rot270_act, rotate_act]
        )
        roi_actions = self.operation_end_actions
        actions = [
            logp1_act,
            flatfield_act,
            None,
            rotate_menu,
            None,
            resize_act,
        ]
        return base_actions + actions + roi_actions

    def create_computing_actions(self):
        """Create computing actions"""
        base_actions = super().create_computing_actions()
        proc = self.processor
        # TODO: [P3] Add "Create ROI grid..." action to create a regular grid or ROIs
        cent_act = self.cra(
            _("Centroid"), proc.compute_centroid, tip=_("Compute image centroid")
        )
        encl_act = self.cra(
            _("Minimum enclosing circle center"),
            proc.compute_enclosing_circle,
            tip=_("Compute smallest enclosing circle center"),
        )
        peak_act = self.cra(
            _("2D peak detection"),
            proc.compute_peak_detection,
            tip=_("Compute automatic 2D peak detection"),
        )
        contour_act = self.cra(
            _("Contour detection"),
            proc.compute_contour_shape,
            tip=_("Compute contour shape fit"),
        )
        self.actlist_1more += [cent_act, encl_act, peak_act, contour_act]
        return base_actions + [cent_act, encl_act, peak_act, contour_act]



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Visualization tools (based on guiqwt)
"""

import guiqwt.image
import guiqwt.plot
import guiqwt.tools
from guidata.configtools import get_icon
from guiqwt.builder import make

from codraft.config import _
from codraft.utils.qthelpers import exec_dialog
from codraft.utils.tests import get_default_test_name

TEST_NB = {}


def get_name_title(name, title):
    """Return (default) widget name and title"""
    if name is None:
        TEST_NB[name] = TEST_NB.setdefault(name, 0) + 1
        name = get_default_test_name(f"{TEST_NB[name]:02d}")
    if title is None:
        title = f'{_("Test dialog")} `{name}`'
    return name, title


def create_curve_dialog(name=None, title=None, xlabel=None, ylabel=None):
    """Create Curve Dialog"""
    name, title = get_name_title(name, title)
    win = guiqwt.plot.CurveDialog(
        edit=False,
        icon=get_icon("codraft.svg"),
        toolbar=True,
        wintitle=title,
        options=dict(xlabel=xlabel, ylabel=ylabel),
    )
    win.setObjectName(name)
    return win


def view_curve_items(items, name=None, title=None, xlabel=None, ylabel=None):
    """Create a curve dialog and plot items"""
    win = create_curve_dialog(name=name, title=title, xlabel=xlabel, ylabel=ylabel)
    plot = win.get_plot()
    for item in items:
        plot.add_item(item)
    exec_dialog(win)


def view_curves(data_or_datalist, name=None, title=None, xlabel=None, ylabel=None):
    """Create a curve dialog and plot curves"""
    if isinstance(data_or_datalist, (tuple, list)):
        datalist = data_or_datalist
    else:
        datalist = [data_or_datalist]
    items = []
    for data in datalist:
        if len(data) == 2:
            xdata, ydata = data
            item = make.mcurve(xdata, ydata)
        else:
            item = make.mcurve(data)
        items.append(item)
    view_curve_items(items, name=name, title=title, xlabel=xlabel, ylabel=ylabel)


def create_image_dialog(name=None, title=None, xlabel=None, ylabel=None):
    """Create Image Dialog"""
    name, title = get_name_title(name, title)
    win = guiqwt.plot.ImageDialog(
        edit=False,
        icon=get_icon("codraft.svg"),
        toolbar=True,
        wintitle=title,
        options=dict(xlabel=xlabel, ylabel=ylabel),
    )
    win.setObjectName(name)
    for toolklass in (
        guiqwt.tools.LabelTool,
        guiqwt.tools.VCursorTool,
        guiqwt.tools.HCursorTool,
        guiqwt.tools.XCursorTool,
        guiqwt.tools.AnnotatedRectangleTool,
        guiqwt.tools.AnnotatedCircleTool,
        guiqwt.tools.AnnotatedEllipseTool,
        guiqwt.tools.AnnotatedSegmentTool,
        guiqwt.tools.AnnotatedPointTool,
    ):
        win.add_tool(toolklass, switch_to_default_tool=True)
    return win


def view_image_items(items, name=None, title=None, xlabel=None, ylabel=None):
    """Create an image dialog and show items"""
    win = create_image_dialog(name=name, title=title, xlabel=xlabel, ylabel=ylabel)
    plot = win.get_plot()
    for item in items:
        plot.add_item(item)
    exec_dialog(win)


def view_images(data_or_datalist, name=None, title=None, xlabel=None, ylabel=None):
    """Create an image dialog and show images"""
    if isinstance(data_or_datalist, (tuple, list)):
        datalist = data_or_datalist
    else:
        datalist = [data_or_datalist]
    items = []
    for data in datalist:
        item = make.image(data, interpolation="nearest", eliminate_outliers=0.1)
        items.append(item)
    view_image_items(items, name=name, title=title, xlabel=xlabel, ylabel=ylabel)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Basic application launcher test 3

- Iterate over all image creation types and data types
- For each image, test all 1->1 computations
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from codraft.core.model.image import (
    Gauss2DParam,
    ImageDatatypes,
    ImageTypes,
    new_image_param,
)
from codraft.tests import codraft_app_context
from codraft.tests.newobject_unit import iterate_image_creation
from codraft.tests.scenario_sig_app import test_compute_11_operations

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Simple test"""
    with codraft_app_context() as win:
        panel = win.imagepanel
        newparam = new_image_param(itype=ImageTypes.GAUSS, dtype=ImageDatatypes.UINT8)
        addparam = Gauss2DParam()
        addparam.x0 = addparam.y0 = 3
        addparam.sigma = 5
        panel.new_object(newparam, addparam=addparam, edit=False)
        panel.processor.compute_wiener()

        for image in iterate_image_creation(500, non_zero=True):
            panel.add_object(image)
            test_compute_11_operations(panel, 0)
            panel.remove_all_objects()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Codra
# Pierre Raybaut

"""
Module providing test utilities
"""

import functools
import os
import os.path as osp
import pathlib
import sys
import tempfile
from contextlib import contextmanager

from guidata.configtools import get_module_data_path

from codraft.env import execenv

TST_PATH = []


def add_test_path_from_env(envvar):
    """Appends test data path from environment variable (fails silently)"""
    path = os.environ.get(envvar)
    if path:
        TST_PATH.append(path)


def add_test_module_path(modname, relpath):
    """
    Appends test data path relative to a module name.
    Used to add module local data that resides in a module directory
    but will be shipped under sys.prefix / share/ ...

    modname must be the name of an already imported module as found in
    sys.modules
    """
    TST_PATH.append(get_module_data_path(modname, relpath=relpath))


def get_test_fnames(pattern):
    """
    Return the absolute path list to test files with specified pattern

    Pattern may be a file name (basename), a wildcard (e.g. *.txt)...
    """
    pathlist = []
    for pth in TST_PATH:
        pathlist += sorted(pathlib.Path(pth).rglob(pattern))
    if not pathlist:
        raise FileNotFoundError(f"Test file(s) {pattern} not found")
    return [str(path) for path in pathlist]


def try_open_test_data(title, pattern):
    """Decorator handling test data opening"""

    def try_open_test_data_decorator(func):
        """Decorator handling test data opening"""

        @functools.wraps(func)
        def func_wrapper():
            """Decorator wrapper function"""
            execenv.print(title + ":")
            execenv.print("-" * len(title))
            try:
                for fname in get_test_fnames(pattern):
                    execenv.print(f"=> Opening: {fname}")
                    func(fname)
            except FileNotFoundError:
                execenv.print(f"  No test data available for {pattern}")
            finally:
                execenv.print(os.linesep)

        return func_wrapper

    return try_open_test_data_decorator


def get_default_test_name(suffix=None):
    """Return default test name based on script name"""
    name = osp.splitext(osp.basename(sys.argv[0]))[0]
    if suffix is not None:
        name += "_" + suffix
    return name


def get_output_data_path(extension, suffix=None):
    """Return full path for data file with extension, generated by a test script"""
    name = get_default_test_name(suffix)
    return osp.join(TST_PATH[0], f"{name}.{extension}")


@contextmanager
def temporary_directory():
    """Create a temporary directory and clean-up afterwards"""
    #  TemporaryDirectory is not used within a "with" statement in order to ignore
    #  errors occuring when cleaning up directory at exit
    #  TODO: [P3] Requires Python 3.10 / Use "ignore_cleanup_errors=True" instead
    #  In other words: this function will be replaced by TemporaryDirectory context mgr
    tmp = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
    try:
        yield tmp.name
    finally:
        try:
            tmp.cleanup()
        except (PermissionError, RecursionError):
            pass



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Log viewer test: raise an exception and create a seg fault in CodraFT
"""


from codraft.core.gui.main import CodraFTMainWindow
from codraft.env import execenv
from codraft.utils.qthelpers import qt_app_context

SHOW = False  # Do not show test in GUI-based test launcher


def error():
    """Raise an exception and create a seg fault in CodraFT"""
    execenv.unattended = True
    with qt_app_context(exec_loop=True):
        win = CodraFTMainWindow()
        win.test_segfault_error()


if __name__ == "__main__":
    error()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Run all tests in unattended mode
"""

import argparse
import os
import os.path as osp

from guidata.guitest import get_tests

import codraft
from codraft import __version__, config
from codraft.utils.tests import TST_PATH

SHOW = True  # Show test in GUI-based test launcher


def get_test_modules(package, contains=""):
    """Return test module list for package"""
    return [
        tmod
        for tmod in get_tests(package)
        if osp.basename(tmod.path) != osp.basename(__file__) and contains in tmod.path
    ]


def run_all_tests(args="", contains="", timeout=None, other_package=None):
    """Run all CodraFT tests"""
    testmodules = get_test_modules(codraft, contains=contains)
    testnb = len(get_tests(codraft)) - 1
    if other_package is not None:
        testmodules += get_test_modules(other_package, contains=contains)
        testnb += len(get_tests(other_package)) - 1
    tnb = len(testmodules)
    print(f"*** CodraFT v{__version__} automatic unit tests ***")
    print("")
    print("Test parameters:")
    print(f"  Selected {tnb} tests ({testnb} total available)")
    if other_package is not None:
        print("  Additional package:")
        print(f"    {other_package.__name__}")
    print("  Test data path:")
    for path in TST_PATH:
        print(f"    {path}")
    print("  Environment:")
    for vname in ("DATA_CODRAFT", "PYTHONPATH", "DEBUG"):
        print(f"    {vname}={os.environ.get(vname, '')}")
    print("")
    print("Please wait while test scripts are executed (a few minutes).")
    print("Only error messages will be printed out (no message = test OK).")
    print("")
    for idx, testmodule in enumerate(testmodules):
        rpath = osp.relpath(testmodule.path, osp.dirname(codraft.__file__))
        print(f"===[{(idx+1):02d}/{tnb:02d}]=== 🍺 Running test [{rpath}]")
        testmodule.run(args=args, timeout=timeout)


def run(other_package=None):
    """Parse arguments and run tests"""
    config.reset()  # Reset configuration (remove configuration file and initialize it)
    parser = argparse.ArgumentParser(description="Run all test in unattended mode")
    parser.add_argument("--contains", default="")
    parser.add_argument("--timeout", type=int, default=240)
    args = parser.parse_args()
    run_all_tests(
        "--mode unattended --verbose quiet",
        args.contains,
        args.timeout,
        other_package,
    )


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)


"""
CodraFT Signal I/O module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import os.path as osp
import re

import numpy as np

from codraft.core.io.conv import data_to_xy
from codraft.core.model.signal import SignalParam, create_signal

# ==============================================================================
# Signal I/O functions
# ==============================================================================

# TODO: Add a mechanism to allow extending features to other file formats

HEADER_KEY = "HEADER"


def read_signal(filename: str) -> SignalParam:
    """Read CSV or NumPy files, return a signal object (`SignalParam` instance)"""
    reducepath = osp.relpath(filename, osp.join(osp.dirname(filename), osp.pardir))
    signal = create_signal(reducepath)
    if osp.splitext(filename)[1] == ".npy":
        xydata = np.load(filename)
    else:
        for delimiter, comments in zip(("\t", ",", " ", ";"), (None, "#")):
            try:
                # Load everything readable (titles are eventually converted as NaNs)
                xydata = np.genfromtxt(
                    filename, delimiter=delimiter, comments=comments, dtype=float
                )
                # Removing lines with NaNs
                xydata = xydata[~np.isnan(xydata).any(axis=1), :]
                # Trying to read X,Y titles
                line0 = delimiter.join([str(val) for val in xydata[0]])
                header = ""
                with open(filename, "r", encoding="utf-8") as fdesc:
                    lines = fdesc.readlines()
                    for rawline in lines:
                        if rawline.startswith(comments):
                            header += rawline
                            continue
                        line = rawline.replace(" ", "")
                        if line == line0:
                            break
                        try:
                            xlabel, ylabel = rawline.split(delimiter)
                            signal.xlabel = xlabel.strip()
                            signal.ylabel = ylabel.strip()
                            # Trying to parse X,Y units
                            pattern = r"([\S ]*) \(([\S]*)\)"  # Matching "Label (unit)"
                            match = re.match(pattern, signal.xlabel)
                            if match is not None:
                                signal.xlabel, signal.xunit = match.groups()
                            match = re.match(pattern, signal.ylabel)
                            if match is not None:
                                signal.ylabel, signal.yunit = match.groups()
                        except ValueError:
                            pass
                        break
                if header:
                    signal.metadata[HEADER_KEY] = header
                break
            except ValueError:
                continue
        else:
            raise ValueError("Unable to open CSV file")
    assert len(xydata.shape) in (1, 2), "Data not supported"
    if len(xydata.shape) == 1:
        signal.set_xydata(np.arange(xydata.size), xydata)
    else:
        x, y, dx, dy = data_to_xy(xydata)
        signal.set_xydata(x, y, dx, dy)
    return signal


def write_signal(obj: SignalParam, filename: str) -> None:
    """Write signal object to CSV or NumPy file"""
    if osp.splitext(filename)[1] == ".npy":
        np.save(filename, obj.xydata.T)
    else:
        xlabel, ylabel = obj.xlabel or "X", obj.ylabel or "Y"
        if obj.xunit:
            xlabel += f" ({obj.xunit})"
        if obj.yunit:
            ylabel += f" ({obj.yunit})"
        delimiter = ","
        np.savetxt(
            filename,
            obj.xydata.T,
            header=delimiter.join([xlabel, ylabel]),
            delimiter=delimiter,
            comments=obj.metadata.get(HEADER_KEY, ""),
        )



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
HDF5 browser test 1

Try and open all HDF5 test data available.
"""


from codraft.env import execenv
from codraft.tests.data import get_test_fnames
from codraft.utils.qthelpers import exec_dialog, qt_app_context
from codraft.widgets.h5browser import H5BrowserDialog

SHOW = True  # Show test in GUI-based test launcher


def h5browser_test(pattern=None):
    """HDF5 browser test"""
    with qt_app_context():
        fnames = get_test_fnames("*.h5" if pattern is None else pattern)
        for index, fname in enumerate(fnames):
            execenv.print(f"Opening: {fname}")
            dlg = H5BrowserDialog(None, size=(1050, 450))
            dlg.setup(fname)
            dlg.browser.tree.toggle_all(True)
            dlg.browser.tree.select_all(True)
            dlg.setObjectName(dlg.objectName() + f"_{index:02d}")
            exec_dialog(dlg)


if __name__ == "__main__":
    h5browser_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
2D peak detection test

Testing the following:
  - Create a test image with multiple peaks
  - Compute 2D peak detection and show points on image
"""

from codraft.config import _
from codraft.core.gui.processor.image import PeakDetectionParam
from codraft.core.model import image as imod
from codraft.tests import codraft_app_context, take_plotwidget_screenshot
from codraft.tests.data import get_peak2d_data

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run 2D peak detection scenario"""
    with codraft_app_context() as win:
        panel = win.imagepanel

        ima = imod.create_image(_("Test image with peaks"), get_peak2d_data())
        panel.add_object(ima)
        param = PeakDetectionParam()
        param.create_rois = True
        panel.processor.compute_peak_detection(param)
        panel.toggle_show_titles(False)
        take_plotwidget_screenshot(panel, "peak2d_test")


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT
=======

CodraFT is a generic signal and image processing software based on Python scientific
libraries (such as NumPy, SciPy or scikit-image) and Qt graphical user interfaces
(thanks to `guidata`_ and `guiqwt`_ libraries).

.. _guidata: https://pypi.python.org/pypi/guidata
.. _guiqwt: https://pypi.python.org/pypi/guiqwt
"""

import os

__version__ = "2.2.1"
__docurl__ = "https://codraft.readthedocs.io/en/latest/"
__homeurl__ = "https://codra-ingenierie-informatique.github.io/CodraFT/"
__supporturl__ = (
    "https://github.com/CODRA-Ingenierie-Informatique/CodraFT/issues/new/choose"
)

os.environ["CODRAFT_VERSION"] = __version__

try:
    import codraft.core.io  # analysis:ignore
    import codraft.patch  # analysis:ignore
except ImportError:
    if not os.environ.get("CODRAFT_DOC"):
        raise

# Dear (Debian, RPM, ...) package makers, please feel free to customize the
# following path to module's data (images) and translations:
DATAPATH = LOCALEPATH = ""



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Basic application launcher test 4

Create an image object from Scikit-image human mitosis sample,
then open CodraFT to show it.
"""


from skimage.data import human_mitosis

from codraft.core.model.image import create_image
from codraft.tests import codraft_app_context

SHOW = False  # Show test in GUI-based test launcher


def test():
    """Dictionnary/List in metadata (de)serialization test"""
    with codraft_app_context(console=False) as win:
        panel = win.imagepanel
        data = human_mitosis()
        image = create_image("Test image with peaks", data)
        panel.add_object(image)
        panel.processor.compute_peak_detection()
        panel.processor.compute_contour_shape()
        panel.processor.extract_roi()


if __name__ == "__main__":
    test()



import configparser as cp
import os
import os.path as osp
import re


def generate_requirement_tables():
    """Generate CodraFT install requirements RST table.
    This table is inserted into 'installation.rst' when
    building documentation"""
    path = osp.dirname(__file__)
    config = cp.ConfigParser()
    config.read(osp.join(path, os.pardir, "setup.cfg"))
    ireq = config["options"]["install_requires"].strip().splitlines(False)
    requirements = [
        ".. list-table::",
        "    :header-rows: 1",
        "",
        "    * - Name",
        "      - Version (min.)",
    ]
    ireq = ["Python>=3.7", "PyQt=5.15"] + ireq
    for req in ireq:
        mod, _comp, ver = re.split("(>=|<=|=|<|>)", req)
        requirements.append("    * - " + mod)
        requirements.append("      - " + ver)
    with open(osp.join(path, "install_requires.txt"), "w") as fdesc:
        fdesc.write("\n".join(requirements))


if __name__ == "__main__":
    generate_requirement_tables()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Computation / Image module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np
import scipy.ndimage as spi
import scipy.ndimage.filters as spf
import scipy.spatial as spt
from numpy import ma
from skimage import measure


def scale_data_to_min_max(data: np.ndarray, zmin, zmax):
    """Scale array `data` to fit [zmin, zmax] dynamic range"""
    dmin = data.min()
    dmax = data.max()
    fdata = np.array(data, dtype=float)
    fdata -= dmin
    fdata *= float(zmax - zmin) / (dmax - dmin)
    fdata += float(zmin)
    return np.array(fdata, data.dtype)


def flatfield(rawdata: np.ndarray, flatdata: np.ndarray, threshold: float = None):
    """Compute flat-field correction"""
    dtemp = np.array(rawdata, dtype=np.float64, copy=True) * flatdata.mean()
    dunif = np.array(flatdata, dtype=np.float64, copy=True)
    dunif[dunif == 0] = 1.0
    dcorr_all = np.array(dtemp / dunif, dtype=rawdata.dtype)
    dcorr = np.array(rawdata, copy=True)
    dcorr[rawdata > threshold] = dcorr_all[rawdata > threshold]
    return dcorr


def get_centroid_fourier(data: np.ndarray):
    """Return image centroid using Fourier algorithm"""
    # Fourier transform method as discussed by Weisshaar et al.
    # (http://www.mnd-umwelttechnik.fh-wiesbaden.de/pig/weisshaar_u5.pdf)
    rows, cols = data.shape
    if rows == 1 or cols == 1:
        return 0, 0

    i = np.arange(0, rows).reshape(1, rows)
    sin_a = np.sin((i - 1) * 2 * np.pi / (rows - 1)).T
    cos_a = np.cos((i - 1) * 2 * np.pi / (rows - 1)).T

    j = np.arange(0, cols).reshape(cols, 1)
    sin_b = np.sin((j - 1) * 2 * np.pi / (cols - 1)).T
    cos_b = np.cos((j - 1) * 2 * np.pi / (cols - 1)).T

    a = (cos_a * data).sum()
    b = (sin_a * data).sum()
    c = (data * cos_b).sum()
    d = (data * sin_b).sum()

    rphi = (0 if b > 0 else 2 * np.pi) if a > 0 else np.pi
    cphi = (0 if d > 0 else 2 * np.pi) if c > 0 else np.pi

    if a * c == 0.0:
        return 0, 0

    row = (np.arctan(b / a) + rphi) * (rows - 1) / (2 * np.pi) + 1
    col = (np.arctan(d / c) + cphi) * (cols - 1) / (2 * np.pi) + 1
    try:
        row = int(row)
    except ma.MaskError:
        row = np.nan
    try:
        col = int(col)
    except ma.MaskError:
        col = np.nan
    return row, col


def get_absolute_level(data: np.ndarray, level: float):
    """Return absolute level"""
    if not isinstance(level, float) or level < 0.0 or level > 1.0:
        raise ValueError("Level must be a float between 0. and 1.")
    return (float(np.nanmin(data)) + float(np.nanmax(data))) * level


def get_enclosing_circle(data: np.ndarray, level: float = 0.5):
    """Return (x, y, radius) for the circle contour enclosing image
    values above threshold relative level (.5 means FWHM)

    Raise ValueError if no contour was found"""
    data_th = data.copy()
    data_th[data <= get_absolute_level(data, level)] = 0.0
    contours = measure.find_contours(data_th)
    model = measure.CircleModel()
    result = None
    max_radius = 1.0
    for contour in contours:
        if model.estimate(contour):
            yc, xc, radius = model.params
            if radius > max_radius:
                result = (int(xc), int(yc), radius)
                max_radius = radius
    if result is None:
        raise ValueError("No contour was found")
    return result


def distance_matrix(coords: list) -> np.ndarray:
    """Return distance matrix from coords"""
    return np.triu(spt.distance.cdist(coords, coords, "euclidean"))


def get_2d_peaks_coords(
    data: np.ndarray, size: int = None, level: float = 0.5
) -> np.ndarray:
    """Detect peaks in image data, return coordinates.

    If neighborhoods size is None, default value is the highest value
    between 50 pixels and the 1/40th of the smallest image dimension.

    Detection threshold level is relative to difference
    between data maximum and minimum values.
    """
    if size is None:
        size = max(min(data.shape) // 40, 50)
    data_max = spf.maximum_filter(data, size)
    data_min = spf.minimum_filter(data, size)
    data_diff = data_max - data_min
    diff = (data_max - data_min) > get_absolute_level(data_diff, level)
    maxima = data == data_max
    maxima[diff == 0] = 0
    labeled, _num_objects = spi.label(maxima)
    slices = spi.find_objects(labeled)
    coords = []
    for dy, dx in slices:
        x_center = int(0.5 * (dx.start + dx.stop - 1))
        y_center = int(0.5 * (dy.start + dy.stop - 1))
        coords.append((x_center, y_center))
    if len(coords) > 1:
        # Eventually removing duplicates
        dist = distance_matrix(coords)
        for index in reversed(np.unique(np.where((dist < size) & (dist > 0))[1])):
            coords.pop(index)
    return np.array(coords)


def get_contour_shapes(
    data: np.ndarray, shape: str = "ellipse", level: float = 0.5
) -> np.ndarray:
    """Find iso-valued contours in a 2D array, above relative level (.5 means FWHM),
    then fit contours with shape ('ellipse' or 'circle')

    Return NumPy array containing coordinates of shapes."""
    # pylint: disable=too-many-locals
    contours = measure.find_contours(data, level=get_absolute_level(data, level))
    coords = []
    for contour in contours:
        if shape == "circle":
            model = measure.CircleModel()
            if model.estimate(contour):
                yc, xc, r = model.params
                if r <= 1.0:
                    continue
                coords.append([xc - r, yc, xc + r, yc])
        elif shape == "ellipse":
            model = measure.EllipseModel()
            if model.estimate(contour):
                yc, xc, b, a, theta = model.params
                if a <= 1.0 or b <= 1.0:
                    continue
                dxa, dya = a * np.cos(theta), a * np.sin(theta)
                dxb, dyb = b * np.sin(theta), b * np.cos(theta)
                x1, y1, x2, y2 = xc - dxa, yc - dya, xc + dxa, yc + dya
                x3, y3, x4, y4 = xc - dxb, yc - dyb, xc + dxb, yc + dyb
                coords.append([x1, y1, x2, y2, x3, y3, x4, y4])
        else:
            raise NotImplementedError(f"Invalid contour model {model}")
    return np.array(coords)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Enclosing circle test

Testing enclsoing circle function on various test images.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from guiqwt.builder import make

from codraft.config import _
from codraft.core.computation.image import get_centroid_fourier, get_enclosing_circle
from codraft.env import execenv
from codraft.tests.data import get_laser_spot_data
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_image_items

SHOW = True  # Show test in GUI-based test launcher


def test_enclosingcircle(data):
    """Enclosing circle test function"""
    items = []
    items += [make.image(data, interpolation="nearest", eliminate_outliers=1.0)]

    # Computing centroid coordinates
    row, col = get_centroid_fourier(data)
    label = _("Centroid") + " (%d, %d)"
    execenv.print(label % (row, col))
    cursor = make.xcursor(col, row, label=label)
    cursor.set_resizable(False)
    cursor.set_movable(False)
    items.append(cursor)

    x, y, radius = get_enclosing_circle(data)
    circle = make.circle(x - radius, y - radius, x + radius, y + radius)
    circle.set_readonly(True)
    circle.set_resizable(False)
    circle.set_movable(False)
    items.append(circle)
    execenv.print(x, y, radius)
    execenv.print("")

    view_image_items(items)


def enclosing_circle_test():
    """Test"""
    with qt_app_context():
        for data in get_laser_spot_data():
            test_enclosingcircle(data)


if __name__ == "__main__":
    enclosing_circle_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT main window status bar widgets
"""


import psutil
from guidata.qthelpers import get_std_icon
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from codraft.config import DEBUG, _
from codraft.env import execenv


class MemoryStatus(QW.QWidget):
    """
    Memory status widget

    :param int threshold: available memory thresold (MB)
    :param int delay: update interval (s)
    :param QWidget parent: parent widget
    """

    SIG_MEMORY_ALARM = QC.Signal(bool)

    def __init__(self, threshold: int = 500, delay: int = 2, parent: QW.QWidget = None):
        super().__init__(parent)
        self.demo_mode = False
        self.ko_icon = get_std_icon("MessageBoxWarning")
        layout = QW.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.icon = QW.QLabel()
        self.label = QW.QLabel()
        layout.addWidget(self.icon)
        layout.addWidget(self.label)
        self.__threshold = threshold * (1024**2)
        self.label.setMinimumWidth(self.label.fontMetrics().width("000%"))
        self.timer = QC.QTimer()
        self.timer.timeout.connect(self.update_status)
        self.timer.start(delay * 1000)
        self.update_status()

    def set_demo_mode(self, state):
        """Set demo mode state (used when taking screenshots)"""
        self.demo_mode = state
        self.update_status()

    def update_status(self):
        """Update status widget"""
        mem = psutil.virtual_memory()
        memok = mem.available > self.__threshold
        self.SIG_MEMORY_ALARM.emit(not memok)
        if DEBUG and not memok:
            execenv.print(f"=== Memory available: {mem.available//(1024**2)} MB ===")
        self.label.setStyleSheet("" if memok else "color: red")
        size = self.label.sizeHint().height()
        self.icon.setPixmap(QG.QPixmap() if memok else self.ko_icon.pixmap(size, size))
        mem_percent = 65 if self.demo_mode else int(mem.percent)
        self.label.setText(_("Memory:") + f" {mem_percent}%")



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Basic application launcher test 2

Create signal and image objects (with circles, rectangles, segments and markers),
then open CodraFT to show them.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from codraft.app import run
from codraft.tests import data as test_data

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Simple test"""
    sig1 = test_data.create_test_signal1()
    sig2 = test_data.create_test_signal2()
    size = 2000
    ima1 = test_data.create_test_image1(size)
    ima2 = test_data.create_test_image2(size, with_annotations=True)
    run(objects=(sig1, sig2, ima1, ima2), size=(1200, 550))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Plot item list classes

These classes handle guiqwt plot items for signal and image panels.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from guiqwt.builder import make
from guiqwt.curve import GridItem
from guiqwt.label import LegendBoxItem
from guiqwt.styles import style_generator

from codraft.config import Conf


class BaseItemList:
    """Object handling plot items associated to objects (signals/images)"""

    def __init__(self, panel, objlist, plotwidget):
        self._enable_cleanup_dataview = True
        self.panel = panel
        self.objlist = objlist
        self.plotwidget = plotwidget
        self.plot = plotwidget.get_plot()
        self.__plotitems = []  # plot items associated to objects (sig/ima)
        self.__shapeitems = []

    def __len__(self):
        """Return number of items"""
        return len(self.__plotitems)

    def __getitem__(self, row):
        """Return item at row"""
        return self.__plotitems[row]

    def __setitem__(self, row, item):
        """Set item at row"""
        self.__plotitems[row] = item

    def __delitem__(self, row):
        """Del item at row"""
        item = self.__plotitems.pop(row)
        self.plot.del_item(item)

    def __iter__(self):
        """Return an iterator over items"""
        yield from self.__plotitems

    def append(self, item):
        """Append item"""
        self.__plotitems.append(item)

    def insert(self, row):
        """Insert object at row index"""
        self.__plotitems.insert(row, None)

    def add_item_to_plot(self, row):
        """Add plot item to plot"""
        item = self.objlist[row].make_item()
        item.set_readonly(True)
        if row < len(self):
            self[row] = item
        else:
            self.append(item)
        self.plot.add_item(item)
        return item

    def make_item_from_existing(self, row):
        """Make plot item from existing object/item at row"""
        return self.objlist[row].make_item(update_from=self[row])

    def update_item(self, row, ref_item=None):
        """Update plot item associated to data"""
        self.objlist[row].update_item(self[row], ref_item=ref_item)

    def add_shapes(self, row):
        """Add geometric shape items associated to computed results and annotations"""
        obj = self.objlist[row]
        if obj.metadata:
            # Performance optimization: block `guiqwt.baseplot.BasePlot` signals,
            # add all items except the last one, unblock signals, then add the last one
            # (this avoids some unnecessary refresh process by guiqwt)
            items = list(obj.iterate_shape_items(editable=False))
            if items:
                block = self.plot.blockSignals(True)
                for item in items[:-1]:
                    self.plot.add_item(item)
                    self.__shapeitems.append(item)
                self.plot.blockSignals(block)
                self.plot.add_item(items[-1])
                self.__shapeitems.append(items[-1])

    def remove_all(self):
        """Remove all plot items"""
        self.__plotitems = []
        self.plot.del_all_items()

    def remove_all_shape_items(self):
        """Remove all geometric shapes associated to result items"""
        if set(self.__shapeitems).issubset(set(self.plot.items)):
            self.plot.del_items(self.__shapeitems)
        self.__shapeitems = []

    def refresh_plot(self, only_row: int = None):
        """Refresh plot (if row is not None, refresh only plot associated to row)"""
        if only_row is None:
            rows = self.objlist.get_selected_rows()
            if self._enable_cleanup_dataview and len(rows) == 1:
                self.cleanup_dataview()
            self.remove_all_shape_items()
            for item in self:
                if item is not None:
                    item.hide()
        else:
            rows = [only_row]
        title_keys = ("title", "xlabel", "ylabel", "zlabel", "xunit", "yunit", "zunit")
        titles_dict = {}
        if rows:
            ref_item = None
            for i_row, row in enumerate(rows):
                for key in title_keys:
                    title = getattr(self.objlist[row], key, "")
                    value = titles_dict.get(key)
                    if value is None:
                        titles_dict[key] = title
                    elif value != title:
                        titles_dict[key] = ""
                item = self[row]
                if item is None:
                    item = self.add_item_to_plot(row)
                else:
                    if i_row == 0:
                        make.style = style_generator()
                    self.update_item(row, ref_item=ref_item)
                    if ref_item is None:
                        ref_item = item
                self.plot.set_item_visible(item, True, replot=False)
                self.plot.set_active_item(item)
                item.unselect()
                self.add_shapes(row)
            self.plot.replot()
        else:
            for key in title_keys:
                titles_dict[key] = ""
        tdict = titles_dict
        tdict["ylabel"] = (tdict["ylabel"], tdict.pop("zlabel"))
        tdict["yunit"] = (tdict["yunit"], tdict.pop("zunit"))
        self.plot.set_titles(**titles_dict)
        self.plot.do_autoscale()

    def toggle_cleanup_dataview(self, state):
        """Toggle clean up data view option"""
        self._enable_cleanup_dataview = state

    def cleanup_dataview(self):
        """Clean up data view"""
        # Performance optimization: using `baseplot.BasePlot.del_items` instead of
        # `baseplot.BasePlot.del_item` (avoid emitting unnecessary signals)
        self.plot.del_items(
            [
                item
                for item in self.plot.items[:]
                if item not in self and not isinstance(item, (LegendBoxItem, GridItem))
            ]
        )

    def get_current_plot_options(self):
        """
        Return standard signal/image plot options

        :return: Dictionary containing plot arguments for CurveDialog/ImageDialog
        """
        return dict(
            xlabel=self.plot.get_axis_title("bottom"),
            ylabel=self.plot.get_axis_title("left"),
            xunit=self.plot.get_axis_unit("bottom"),
            yunit=self.plot.get_axis_unit("left"),
        )


class SignalItemList(BaseItemList):
    """Object handling signal plot items, plot dialogs, plot options"""

    # Nothing specific to signals, as of today


class ImageItemList(BaseItemList):
    """Object handling image plot items, plot dialogs, plot options"""

    def refresh_plot(self, only_row: int = None):
        """Refresh plot (if row is not None, refresh only plot associated to row)"""
        super().refresh_plot(only_row)
        self.plotwidget.contrast.setVisible(Conf.view.show_contrast.get(True))

    def cleanup_dataview(self):
        """Clean up data view"""
        for widget in (self.plotwidget.xcsw, self.plotwidget.ycsw):
            widget.hide()
        super().cleanup_dataview()

    def get_current_plot_options(self):
        """
        Return standard signal/image plot options

        :return: Dictionary containing plot arguments for CurveDialog/ImageDialog
        """
        options = super().get_current_plot_options()
        options.update(
            dict(
                zlabel=self.plot.get_axis_title("right"),
                zunit=self.plot.get_axis_unit("right"),
                show_contrast=True,
            )
        )
        return options



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
ROI editor widgets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc

import numpy as np
from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action
from guiqwt.annotations import AnnotatedCircle
from guiqwt.builder import make
from guiqwt.interfaces import IImageItemType
from guiqwt.label import ObjectInfo
from qtpy import QtWidgets as QW

from codraft.config import Conf, _
from codraft.core.model.base import ObjectItf
from codraft.core.model.image import RoiDataGeometries


class ROIEditorData:
    """ROI Editor data"""

    def __init__(self, roidata: np.ndarray = None, singleobj: bool = None):
        self.__singleobj = None
        if roidata is not None:
            roidata = np.array(roidata, dtype=int)
        self.roidata = roidata
        self.singleobj = singleobj
        self.modified = None

    @property
    def is_empty(self) -> bool:
        """Return True if there is no ROI"""
        return self.roidata is None or self.roidata.size == 0

    @property
    def singleobj(self) -> bool:
        """Return singleobj parameter"""
        return self.__singleobj

    @singleobj.setter
    def singleobj(self, value: bool):
        """Set singleobj parameter"""
        if value is None:
            value = Conf.proc.extract_roi_singleobj.get(False)
        self.__singleobj = value
        Conf.proc.extract_roi_singleobj.set(value)


class BaseROIEditorMeta(type(QW.QWidget), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class BaseROIEditor(QW.QWidget, metaclass=BaseROIEditorMeta):
    """ROI Editor"""

    ICON_NAME = None
    OBJ_NAME = None

    def __init__(
        self, parent: QW.QDialog, obj: ObjectItf, extract: bool, singleobj: bool = None
    ):
        super().__init__(parent)
        parent.accepted.connect(self.dialog_accepted)
        self.plot = parent.get_plot()
        self.obj = obj
        self.extract = extract
        self.__modified = None

        self.__data = ROIEditorData(singleobj=singleobj)

        self.fmt = obj.metadata.get(obj.METADATA_FMT, "%s")
        self.roi_items = list(obj.iterate_roi_items(self.fmt, True))

        for roi_item in self.roi_items:
            self.plot.add_item(roi_item)
            self.plot.set_active_item(roi_item)

        self.add_btn = None
        self.singleobj_btn = None
        self.setup_widget()

        self.update_roi_titles()
        self.plot.SIG_ITEMS_CHANGED.connect(lambda _plt: self.update_roi_titles())
        self.plot.SIG_ITEM_REMOVED.connect(self.item_removed)
        self.plot.SIG_RANGE_CHANGED.connect(lambda _rng, _min, _max: self.item_moved())
        self.plot.SIG_ANNOTATION_CHANGED.connect(lambda _plt: self.item_moved())

        #  In "extract mode", the dialog box OK button should always been enabled
        #  when at least one ROI is defined,
        #  whereas in non-extract mode (when editing ROIs) the OK button is by default
        #  disabled (until ROI data is modified)
        self.modified = extract

    @property
    def modified(self) -> bool:
        """Return dialog modified state"""
        return self.__modified

    @modified.setter
    def modified(self, value: bool):
        """Set dialog modified state"""
        self.__modified = value
        dlg = self.parent()
        if self.extract:
            #  In "extract mode", OK button is enabled when at least one ROI is defined
            value = value and len(self.roi_items) > 0
        dlg.button_box.button(QW.QDialogButtonBox.Ok).setEnabled(value)

    def dialog_accepted(self):
        """Parent dialog was accepted: updating ROI Editor data"""
        coords = []
        for roi_item in self.roi_items:
            coords.append(list(self.get_roi_item_coords(roi_item)))
        self.__data.roidata = self.obj.roi_coords_to_indexes(coords)
        if self.singleobj_btn is not None:
            self.__data.singleobj = self.singleobj_btn.isChecked()
        self.__data.modified = self.modified

    def get_data(self) -> ROIEditorData:
        """Get ROI Editor data (results of the dialog box)"""
        return self.__data

    def setup_widget(self):
        """Setup ROI editor widget"""
        self.add_btn = QW.QPushButton(
            get_icon(self.ICON_NAME), _("Add region of interest"), self
        )
        layout = QW.QHBoxLayout()
        layout.addWidget(self.add_btn)
        if self.extract:
            self.singleobj_btn = QW.QCheckBox(
                _("Extract all regions of interest into a single %s object")
                % self.OBJ_NAME,
                self,
            )
            layout.addWidget(self.singleobj_btn)
            self.singleobj_btn.setChecked(self.__data.singleobj)
        layout.addStretch()
        self.setLayout(layout)

    def add_roi_item(self, roi_item):
        """Add ROI item to plot and refresh titles"""
        self.plot.unselect_all()
        self.roi_items.append(roi_item)
        self.update_roi_titles()
        self.modified = True
        self.plot.add_item(roi_item)
        self.plot.set_active_item(roi_item)

    @abc.abstractmethod
    def update_roi_titles(self):
        """Update ROI annotation titles"""

    def item_removed(self, item):
        """Item was removed. Since all items are read-only except ROIs...
        this must be an ROI."""
        assert item in self.roi_items
        self.roi_items.remove(item)
        self.modified = True
        self.update_roi_titles()

    def item_moved(self):
        """ROI plot item has just been moved"""
        self.modified = True

    @staticmethod
    @abc.abstractmethod
    def get_roi_item_coords(roi_item):
        """Return ROI item coords"""


class ROIRangeInfo(ObjectInfo):
    """ObjectInfo for ROI selection"""

    def __init__(self, roi_items):
        self.roi_items = roi_items

    def get_text(self):
        textlist = []
        for index, roi_item in enumerate(self.roi_items):
            x0, x1 = roi_item.get_range()
            textlist.append(f"ROI{index:02d}: {x0} ≤ x ≤ {x1}")
        return "<br>".join(textlist)


class SignalROIEditor(BaseROIEditor):
    """Signal ROI Editor"""

    ICON_NAME = "signal_roi_new.svg"
    OBJ_NAME = _("signal")

    def setup_widget(self):
        """Setup ROI editor widget"""
        super().setup_widget()
        info = ROIRangeInfo(self.roi_items)
        info_label = make.info_label("BL", info, title=_("Regions of interest"))
        self.plot.add_item(info_label)
        self.info_label = info_label
        self.add_btn.clicked.connect(self.add_roi)

    def add_roi(self):
        """Simply add an ROI"""
        roi_item = self.obj.new_roi_item(self.fmt, True, editable=True)
        self.add_roi_item(roi_item)

    def update_roi_titles(self):
        """Update ROI annotation titles"""
        super().update_roi_titles()
        self.info_label.update_text()

    @staticmethod
    def get_roi_item_coords(roi_item):
        """Return ROI item coords"""
        return roi_item.get_range()


class ImageROIEditor(BaseROIEditor):
    """Image ROI Editor"""

    ICON_NAME = "image_roi_new.svg"
    OBJ_NAME = _("image")

    def setup_widget(self):
        """Setup ROI editor widget"""
        super().setup_widget()
        item = self.plot.get_items(item_type=IImageItemType)[0]
        item.set_mask_visible(False)
        menu = QW.QMenu()
        rectact = create_action(
            self,
            _("Rectangular ROI"),
            lambda: self.add_roi(RoiDataGeometries.RECTANGLE),
            icon=get_icon("rectangle.png"),
        )
        circact = create_action(
            self,
            _("Circular ROI"),
            lambda: self.add_roi(RoiDataGeometries.CIRCLE),
            icon=get_icon("circle.png"),
        )
        add_actions(menu, (rectact, circact))
        self.add_btn.setMenu(menu)

    def add_roi(self, geometry: RoiDataGeometries):
        """Add new ROI"""
        item = self.obj.new_roi_item(self.fmt, True, editable=True, geometry=geometry)
        self.add_roi_item(item)

    def update_roi_titles(self):
        """Update ROI annotation titles"""
        super().update_roi_titles()
        for index, roi_item in enumerate(self.roi_items):
            roi_item.annotationparam.title = f"ROI{index:02d}"
            roi_item.annotationparam.update_annotation(roi_item)

    @staticmethod
    def get_roi_item_coords(roi_item):
        """Return ROI item coords"""
        x0, y0, x1, y1 = roi_item.get_rect()
        if isinstance(roi_item, AnnotatedCircle):
            y0 = y1 = 0.5 * (y0 + y1)
        return x0, y0, x1, y1



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image tools application test:

  - Testing `ZAxisLogTool`
  - Testing `to_codraft` function (image cross section -> curve)
"""

from guiqwt.baseplot import axes_to_canvas
from guiqwt.tools import CrossSectionTool
from qtpy import QtCore as QC

from codraft.patch import ZAxisLogTool, to_codraft
from codraft.tests import codraft_app_context
from codraft.tests.data import create_test_image3
from codraft.utils.qthelpers import qt_wait

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run image tools test scenario"""
    with codraft_app_context() as win:
        panel = win.imagepanel
        size = 200
        ima = create_test_image3(size)
        panel.add_object(ima)
        plotwidget = panel.itmlist.plotwidget

        # === Testing "ZAxisLogTool" ---------------------------------------------------
        lstool = plotwidget.get_tool(ZAxisLogTool)
        qt_wait(1, except_unattended=True)
        for _index in range(2):
            lstool.activate()
            qt_wait(1, except_unattended=True)

        # === Testing "to_codraft" -----------------------------------------------------
        plot = plotwidget.plot
        cstool = plotwidget.get_tool(CrossSectionTool)
        cstool.activate()
        pos = QC.QPointF(*axes_to_canvas(plot.active_item, size // 2, size // 2))
        cstool.end_rect(plotwidget.plot.filter, pos, pos)
        for csw in (plotwidget.xcsw, plotwidget.ycsw):
            to_codraft(csw.cs_plot)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)


"""
CodraFT HDF5 importer module
"""

# Registering dynamic I/O features:
from codraft.core.io.h5 import generic, mos07636  # pylint: disable=W0611
from codraft.core.io.h5.common import H5Importer  # pylint: disable=W0611



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Result shapes application test:

  - Create an image with metadata shapes and ROI
  - Further tests to be done manually: check if copy/paste metadata works
"""

import numpy as np

from codraft.core.gui.processor.signal import FWHMParam
from codraft.core.model.image import create_image
from codraft.tests import codraft_app_context
from codraft.tests import data as test_data

SHOW = True  # Show test in GUI-based test launcher


def create_image_with_resultshapes():
    """Create test image with resultshapes"""
    data = test_data.create_2d_gaussian(600, np.uint16, x0=2.0, y0=3.0)
    image = create_image("Test image with metadata", data)
    for mshape in test_data.create_resultshapes():
        mshape.add_to(image)
    return image


def test():
    """Result shapes test"""
    obj1 = test_data.create_test_image1()
    obj2 = create_image_with_resultshapes()
    obj2.roi = np.array([[10, 10, 60, 400]], int)
    with codraft_app_context(console=False) as win:
        panel = win.signalpanel
        for noised in (False, True):
            sig = test_data.create_test_signal2(noised=noised)
            panel.add_object(sig)
            panel.processor.compute_fwhm(FWHMParam())
            panel.processor.compute_fw1e2()
        panel.objlist.select_rows((0, 1))
        panel.show_results()
        win.switch_to_image_panel()
        panel = win.imagepanel
        for obj in (obj1, obj2):
            panel.add_object(obj)
        panel.show_results()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT mainwindow automation test:

  - Saving current project (h5 file)
  - Removing all objects
  - Opening another project (h5 file)
  - Access current image metadata
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import os.path as osp

from codraft.core.gui.processor.base import ClipParam
from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.tests.data import create_test_image2, create_test_signal1
from codraft.utils.tests import temporary_directory

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run ROI unit test scenario"""
    with temporary_directory() as tmpdir:
        size = 500
        with codraft_app_context(console=False) as win:
            # === Creating two test signals
            panel = win.signalpanel
            sig1 = create_test_signal1(size)
            panel.add_object(sig1)
            panel.processor.compute_derivative()
            # === Creating two test images
            panel = win.imagepanel
            ima1 = create_test_image2(size, with_annotations=True)
            panel.add_object(ima1)
            param = ClipParam()
            param.value = ima1.data.mean()
            panel.processor.compute_clip(param)
            # === Saving project
            fname = osp.join(tmpdir, "test.h5")
            win.save_to_h5_file(fname)
            # === Removing all objects
            for panel in win.panels:
                panel.remove_all_objects()
            # === Reopening previously saved project
            win.open_h5_files([fname], import_all=True, reset_all=True)
            # === Accessing object metadata
            obj = panel.objlist[1]
            execenv.print(f"Image '{obj.title}':")
            for key, value in obj.metadata.items():
                execenv.print(f'  metadata["{key}"] = {value}')


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Module for taking CodraFT screenshots
"""

from codraft import config
from codraft.app import create
from codraft.tests.data import create_test_image1, create_test_signal1
from codraft.utils.qthelpers import qt_app_context


def take_menu_screenshots():
    """Run the CodraFT application and take screenshots"""
    config.reset()  # Reset configuration (remove configuration file and initialize it)
    sig1 = create_test_signal1()
    ima1 = create_test_image1()
    objects = (sig1, ima1)
    with qt_app_context():
        window = create(splash=False, objects=objects, size=(800, 300))
        window.take_menu_screenshots()
        window.take_screenshot("i_simple_example")
        window.switch_to_signal_panel()
        window.take_screenshot("s_simple_example")
        window.set_modified(False)
        window.close()


if __name__ == "__main__":
    take_menu_screenshots()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT core.gui module

This module handles all GUI features which are specific to CodraFT:

  * core.gui.main: handles CodraFT main window which relies on signal and image panels
  * core.gui.panel: handles CodraFT signal and image panels, relying on:
    * core.gui.actionhandler
    * core.gui.objectlist
    * core.gui.plotitemlist
    * core.gui.roieditor
    * core.gui.processor
  * core.gui.docks: handles CodraFT dockwidgets
  * core.gui.h5io: handles HDF5 browser widget and related features
"""



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Application launcher test 1

Running application a few times in a row with different entry parameters.
"""

from codraft import app
from codraft.env import execenv
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.tests import get_test_fnames

SHOW = True  # Show test in GUI-based test launcher


def test(pattern=None):
    """Testing CodraFT app launcher"""
    if pattern is None:
        pattern = "*.h5"
    execenv.print("HDF5 import test scenario:")
    execenv.print("[1] Loading all h5 files at once")
    with qt_app_context(exec_loop=True):
        app.create(h5files=get_test_fnames(pattern))
    execenv.print("[2] Loading h5 files one by one (only the first 3 files)")
    for fname in get_test_fnames(pattern)[:3]:
        with qt_app_context(exec_loop=True):
            execenv.print(f"      Opening: {fname}")
            app.create(h5files=[fname])


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image contrast test on problematic data

Emphasizing contrast calculation issue when data contains outlier points (e.g. hot
spots) thus preventing from cleaning up contrast histogram or showing valid Z-axis.
"""

import numpy as np

from codraft.utils.qthelpers import qt_app_context
from codraft.utils.tests import get_test_fnames
from codraft.utils.vistools import view_images

SHOW = True  # Show test in GUI-based test launcher


def contrast_test():
    """Contrats test"""
    with qt_app_context():
        view_images(np.load(get_test_fnames("contrast_test_data.npy")[0]))


if __name__ == "__main__":
    contrast_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Miscelleneous utilities
"""

import numpy as np


def to_string(obj):
    """Convert to string, trying utf-8 then latin-1 codec"""
    if isinstance(obj, bytes):
        try:
            return obj.decode()
        except UnicodeDecodeError:
            return obj.decode("latin-1")
    try:
        return str(obj)
    except UnicodeDecodeError:
        return str(obj, encoding="latin-1")


def is_integer_dtype(dtype):
    """Return True if data type is an integer type"""
    return issubclass(np.dtype(dtype).type, np.integer)


def is_complex_dtype(dtype):
    """Return True if data type is a complex type"""
    return issubclass(np.dtype(dtype).type, complex)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Computation / Signal module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np


# ----- Filtering functions ----------------------------------------------------
def moving_average(y, n):
    """Compute moving average"""
    y_padded = np.pad(y, (n // 2, n - 1 - n // 2), mode="edge")
    return np.convolve(y_padded, np.ones((n,)) / n, mode="valid")


# ----- Misc. functions --------------------------------------------------------
def derivative(x, y):
    """Compute numerical derivative"""
    dy = np.zeros_like(y)
    dy[0:-1] = np.diff(y) / np.diff(x)
    dy[-1] = (y[-1] - y[-2]) / (x[-1] - x[-2])
    return dy


def normalize(yin, parameter="maximum"):
    """
    Normalize input array *yin* with respect to parameter *parameter*

    Support values for *parameter*:
        'maximum' (default), 'amplitude', 'sum', 'energy'
    """
    axis = len(yin.shape) - 1
    if parameter == "maximum":
        maximum = np.max(yin, axis)
        if axis == 1:
            maximum = maximum.reshape((len(maximum), 1))
        maxarray = np.tile(maximum, yin.shape[axis]).reshape(yin.shape)
        return yin / maxarray
    if parameter == "amplitude":
        ytemp = np.array(yin, copy=True)
        minimum = np.min(yin, axis)
        if axis == 1:
            minimum = minimum.reshape((len(minimum), 1))
        ytemp -= minimum
        return normalize(ytemp, parameter="maximum")
    if parameter == "sum":
        return yin / yin.sum()
    if parameter == "energy":
        return yin / (yin * yin.conjugate()).sum()
    raise RuntimeError(f"Unsupported parameter {parameter}")


def xy_fft(x, y, shift=True):
    """Compute FFT on X,Y data.

    Args:
        x (np.ndarray): X data
        y (np.ndarray): Y data
        shift (bool, optional): Shift the zero frequency to the center of the spectrum.
            Defaults to True.

    Returns:
        tuple[np.ndarray, np.ndarray]: X,Y data
    """
    y1 = np.fft.fft(y)
    x1 = np.fft.fftfreq(x.shape[-1], d=x[1] - x[0])
    if shift:
        x1 = np.fft.fftshift(x1)
        y1 = np.fft.fftshift(y1)
    return x1, y1


def xy_ifft(x, y, shift=True):
    """Compute iFFT on X,Y data.

    Args:
        x (np.ndarray): X data
        y (np.ndarray): Y data
        shift (bool, optional): Shift the zero frequency to the center of the spectrum.
            Defaults to True.

    Returns:
        tuple[np.ndarray, np.ndarray]: X,Y data
    """
    x1 = np.fft.fftfreq(x.shape[-1], d=x[1] - x[0])
    if shift:
        x1 = np.fft.ifftshift(x1)
        y = np.fft.ifftshift(y)
    y1 = np.fft.ifft(y)
    return x1, y1.real


# ----- Peak detection functions -----------------------------------------------
def peak_indexes(y, thres=0.3, min_dist=1, thres_abs=False):
    #  Copyright (c) 2014 Lucas Hermann Negri
    #  Unmodified code snippet from PeakUtils 1.3.0
    """Peak detection routine.

    Finds the numeric index of the peaks in *y* by taking its first order
    difference. By using *thres* and *min_dist* parameters, it is possible
    to reduce the number of detected peaks. *y* must be signed.

    Parameters
    ----------
    y : ndarray (signed)
        1D amplitude data to search for peaks.
    thres : float between [0., 1.]
        Normalized threshold. Only the peaks with amplitude higher than the
        threshold will be detected.
    min_dist : int
        Minimum distance between each detected peak. The peak with the highest
        amplitude is preferred to satisfy this constraint.
    thres_abs: boolean
        If True, the thres value will be interpreted as an absolute value,
        instead of a normalized threshold.

    Returns
    -------
    ndarray
        Array containing the numeric indexes of the peaks that were detected
    """
    if isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.unsignedinteger):
        raise ValueError("y must be signed")

    if not thres_abs:
        thres = thres * (np.max(y) - np.min(y)) + np.min(y)

    min_dist = int(min_dist)

    # compute first order difference
    dy = np.diff(y)

    # propagate left and right values successively to fill all plateau pixels
    # (0-value)
    (zeros,) = np.where(dy == 0)

    # check if the signal is totally flat
    if len(zeros) == len(y) - 1:
        return np.array([])

    if len(zeros):
        # compute first order difference of zero indexes
        zeros_diff = np.diff(zeros)
        # check when zeros are not chained together
        (zeros_diff_not_one,) = np.add(np.where(zeros_diff != 1), 1)
        # make an array of the chained zero indexes
        zero_plateaus = np.split(zeros, zeros_diff_not_one)

        # fix if leftmost value in dy is zero
        if zero_plateaus[0][0] == 0:
            dy[zero_plateaus[0]] = dy[zero_plateaus[0][-1] + 1]
            zero_plateaus.pop(0)

        # fix if rightmost value of dy is zero
        if len(zero_plateaus) > 0 and zero_plateaus[-1][-1] == len(dy) - 1:
            dy[zero_plateaus[-1]] = dy[zero_plateaus[-1][0] - 1]
            zero_plateaus.pop(-1)

        # for each chain of zero indexes
        for plateau in zero_plateaus:
            median = np.median(plateau)
            # set leftmost values to leftmost non zero values
            dy[plateau[plateau < median]] = dy[plateau[0] - 1]
            # set rightmost and middle values to rightmost non zero values
            dy[plateau[plateau >= median]] = dy[plateau[-1] + 1]

    # find the peaks by using the first order difference
    peaks = np.where(
        (np.hstack([dy, 0.0]) < 0.0)
        & (np.hstack([0.0, dy]) > 0.0)
        & (np.greater(y, thres))
    )[0]

    # handle multiple peaks, respecting the minimum distance
    if peaks.size > 1 and min_dist > 1:
        highest = peaks[np.argsort(y[peaks])][::-1]
        rem = np.ones(y.size, dtype=bool)
        rem[peaks] = False

        for peak in highest:
            if not rem[peak]:
                sl = slice(max(0, peak - min_dist), peak + min_dist + 1)
                rem[sl] = True
                rem[peak] = False

        peaks = np.arange(y.size)[~rem]

    return peaks


def xpeak(x, y):
    """Return default peak X-position (assuming a single peak)"""
    peaks = peak_indexes(y)
    if peaks.size == 1:
        return x[peaks[0]]
    return np.average(x, weights=y)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Annotations application test:

  - Create an image with annotations and ROI
  - Further tests to be done manually: edit "Annotations" and check that
    modifications are taken into account, without affecting the existing ROI
"""
import numpy as np

from codraft.app import run
from codraft.tests import data as test_data

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Annotations test"""
    obj1 = test_data.create_test_image1()
    obj2 = test_data.create_image_with_annotations()
    obj2.roi = np.array([[10, 10, 60, 400]], int)
    run(console=False, objects=(obj1, obj2), size=(1200, 550))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image tools test

Simple image dialog for testing all image tools available in CodraFT
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np
from guiqwt.builder import make

from codraft.tests.data import create_2d_gaussian, create_2d_random
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_image_items

SHOW = True  # Show test in GUI-based test launcher


def image_tools_test(size=2000, dtype=np.uint16):
    """Image tools test"""
    with qt_app_context():
        znoise = create_2d_random(size, dtype)
        zgauss = create_2d_gaussian(size, dtype, x0=5.0, y0=-3.0)
        data = zgauss + znoise
        items = [make.image(data, interpolation="nearest", eliminate_outliers=2.0)]
        view_image_items(items)


if __name__ == "__main__":
    image_tools_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT launcher module
"""

from guidata.configtools import get_image_file_path
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from codraft.config import Conf
from codraft.core.gui.main import CodraFTMainWindow
from codraft.env import execenv
from codraft.utils.qthelpers import qt_app_context


def create(
    splash: bool = True, console: bool = None, objects=None, h5files=None, size=None
) -> CodraFTMainWindow:
    """Create CodraFT application and return mainwindow instance"""
    if splash:
        # Showing splash screen
        pixmap = QG.QPixmap(get_image_file_path("codraft_titleicon.png"))
        splashscreen = QW.QSplashScreen(pixmap, QC.Qt.WindowStaysOnTopHint)
        splashscreen.show()
    window = CodraFTMainWindow(console=console)
    if size is not None:
        width, height = size
        window.resize(width, height)
    if splash:
        splashscreen.finish(window)
    if Conf.main.window_maximized.get(None):
        window.showMaximized()
    else:
        window.showNormal()
    if h5files is not None:
        window.open_h5_files(h5files, import_all=True)
    if objects is not None:
        for obj in objects:
            window.add_object(obj)
    if execenv.h5browser_file is not None:
        window.import_h5_file(execenv.h5browser_file)
    return window


def run(console=None, objects=None, h5files=None, size=None):
    """Run the CodraFT application

    Note: this function is an entry point in `setup.py` and therefore
    may not be moved without modifying the package setup script."""
    if execenv.h5files:
        h5files = ([] if h5files is None else h5files) + execenv.h5files

    with qt_app_context(exec_loop=True):
        window = create(
            splash=True, console=console, objects=objects, h5files=h5files, size=size
        )
        QW.QApplication.processEvents()
        window.check_stable_release()
        window.check_dependencies()
        window.check_for_previous_crash()


if __name__ == "__main__":
    run()
